C     PROGRAM XXGTARMA: DRIVER ROUTINE GTARMA FOR THE ESTIMATION OF THE
C     k-REGIMES THRESHOLD AUTOREGRESSIVE MOVING AVERAGE SUBSET LAG MODEL
C     * ONLY TWO-REGIMES MODELS ARE SUPPORTED
C     THE ALGORITHM IS THE SAME USED IN THE PROGRAM XGATARMA.FOR
C     PROGRAM XGATARMA IS EMBEDDED IN A MONTE CARLO SIMULATION.
C     UNLIKE THE SAVED (ZGATARMA.FOR) EARLY PROGRAM, HERE THE THRESHOLD
C     MODEL USES DIFFERENT RESIDUAL TERMS IN EACH REGIME
C     AT LEAST MSPAN OBSERVATIONS FOR REGIME ARE REQUIRED
C     OTHER ADJUSTMENTS ARE INTRODUCED, THAT IS THE MAX NP NQ MAY DIFFER 
C     FROM THEIR BINARY REPRESENTATION 
C     UNLIKE THE SAVED (YGATARMA.FOR) EARLY PROGRAM, HERE THE GENETIC 
C     ALGORITHM TRIES ALL POSSIBLE VALUES OF THE DELAY PARAMETER d FROM
C     1 TO THE MAXIMUM PRE-SPECIFIED D (dmax IN THE PROGRAM) AS A 
C     CONSEQUENCE, MBIT IS SET TO ITS OWN VALUE AND d IS NOT PROCESSED
C     BY THE GENETIC ALGORITHM, BUT IT IS MAINTAINED IN THE CHROMOSOME
C     FOR THE PROGRAM STRUCTURE MAY REMAIN APPROXIMATELY THE SAME. 
C     THE OTHER MAIN DIFFERENCE RESIDES IN THAT A DIFFERENT NUMBER OF
C     BITS, THAT IS DIFFERENT LENGTH, IS ALLOWED FOR THE NUMBER OF AR -
C     MA PARAMETERS (nbit1 IN THE PROGRAM) AND THE LAG (nbit2 IN THE 
C     PROGRAM). LAST, THE CHOICE BETWEEN THE AIC AND THE BIC IS ACTIVE.
C     AS FOR THE REST, THE PROCEDURE DID NOT CHANGE. PARAMETERS ARE: 
C    * THE NUMBER OF REGIMES k AND THRESHOLD CONSTANTS r1..rk1, k1 = k-1
C    * THE NUMBER np OF AR LAGS AND THE NUMBER nq OF MA LAGS (2^nbit1-1)
C    * THE AR LAGS p(1)..p(np) AND THE MA LAGS q(1)..q(nq) (2^nbit2-1)
C     A NEW FEATURE HAS BEEN ADDED: IF THE FITNESS FUNCTION DOES NOT
C     CHANGE FOR A PRE-SPECIFIED NUMBER OF ITERATIONS, THEN CROSSOVER
C     AND MUTATION PROBABILITIES SWITCH TO NEW PRE-SPECIFIED VALUES.
C    -> TIME SERIES Z, NOB OBSERVATIONS, ARE REPLICATED NREPL TIMES
C    -> THERE IS THE COMMON BLOCK /RAND/ (UNIFORM RANDOM ROUTINE FOR GA)
C    -> set_uniform(iseed1, iseed2) and ppnd16 provide artificial series
                                                                        
      PARAMETER(MOB=5500, MK= 05, MDYP =05, MLAGH=36, NMX=100
     +, MBIT1=2, MBIT2=3, MBIT = MBIT1 + MBIT2 
     +, MP=2**MBIT1-1, MQ=2**MBIT1-1, MX=2*(MP+MQ)+1, MV=1+MP+MQ 
     +, MPLAG=2**MBIT2-1, MQLAG=2**MBIT2-1, MXLAG=1+MPLAG+MQLAG
     +, MAXL=MBIT+MOB+MK*(MBIT1+MP*MBIT2+MBIT1+MQ*MBIT2))
      IMPLICIT REAL*8(A - H , O - Z)
      CHARACTER*72  PNAME, FNAME

C     TIME SERIES DIMENSION

      DIMENSION Z(MOB), X(MOB), KKHN(MOB), XMAT(MOB, MX), PHI(MX)
     +, AUX1(MX), VC1(MX, MX), BB(MX), CC(MX), IP(MX), IQ(MX)
     +, RES(MOB), VFIT(MOB), RHO(MLAGH), COV(MLAGH), etilda(mob)
     +, FZ(MOB), FITX(MOB), FITY(MOB), KKN(MOB), KNG(MOB), KKKN(MOB)
     +, NHP(MK), nxp(mk), LAHP(mk, MPLAG), nxq(mk), LAHQ(mk, MQLAG)
     +, XPHI(MK,MXLAG), XTHETA(MK,MXLAG), XPI(MK,MXLAG)
     +, YPHI(MK,MXLAG), YTHETA(MK,MXLAG), YPI(MK,MXLAG)
     +, PXHP(MXLAG), PXHQ(MXLAG), PXHPI(MXLAG) 

C     THRESHOLD PARAMETERS DIMENSION

      DIMENSION R(MK), NJ(MK), S2(MK)

C     GENETIC ALGORITHM DIMENSION

      DIMENSION NO(NMX, MAXL), NN(NMX, MAXL), LBEST(MAXL), LCHROM(MAXL)
     +,  LCHRO1(MAXL), LCHRO2(MAXL),  LCHRO3(MAXL), LCHRO4(MAXL)
     +,   naux1(mbit), naux2(mx), fitnn(nmx), LKYROM(MAXL)  

C     MONTE CARLO EXPERIMENT DIMENSION

      DIMENSION WNP(MDYP,MK,MK,MP),WNPLAG(MDYP,MK,MP,MK,MPLAG)
     +,WNPP(MDYP,MK,MP,MK,MPLAG),WNPP2(MDYP,MK,MP,MK,MPLAG)
     +,WNQ(MDYP,MK,MK,MQ),WNQLAG(MDYP,MK,MQ,MK,MQLAG)
     +,WNQQ(MDYP,MK,MQ,MK,MQLAG),WNQQ2(MDYP,MK,MQ,MK,MQLAG)
     +,WOBS(MDYP,MK,MK),WOBS2(MDYP,MK,MK)
     +,WTHRE(MDYP,MK,MK),WTHRE2(MDYP,MK,MK)
     +,WCONTA(MDYP,MK),WNK(MDYP,MK),WNDYP(MDYP,MK)
     +,WVSSQ(MDYP,MK),WVSSQ2(MDYP,MK) 

C     k = no. of regimes, k1 = k-1 = no. of threshold constants r(.)
C
C             1                2               ...             k     
C     _________________|_________________|____ ...  ____|_______________
C                     r(1)              r(2)           r(k-1)
C      z(t-d) < r(1)    r(1)<=z(t-d)<r(2)               z(t-d) >= r(k-1)
C
C     icri = 1 Akaike's AIC criterion; icri = 2 Schwartz's BIC criterion
C  Akaike, H. (1974) A new look at the statistical model identification,
C     IEEE Trans. Automatic Control, AC-19, 716-723, 1974.
C  Schwartz, G. (1978) Estimating the dimension of a model, Ann.Statist.
C     6, 461-464, 1978.

      COMMON /PPPP/ PPP(5,13)
      COMMON /RAND/ IX, IY, IZ
      DATA  TOL/ 8.43D-37/
      iseed1 =  531919
      iseed2 =  531977
c  iseed1 = prevprime(531971); iseed2 = nextprime(531971);       
	call set_uniform(iseed1, iseed2)

      LDUMP = 10
      LPRINT  =  6
      iprint  =  0

      OPEN(UNIT=5, FILE='xxgtarma.txt')

      READ(5,*)   NOB, NAC, MSPAN, NK, NDYP, ICRI
      READ(5,  *)  NBIT1, NBIT2
      READ(5, *)   NHP0, NHP1, NP, NPLAG, NQ, NQLAG
      READ(5, *)  NL, P2, P1, MAXIT, XETA
      READ(5, *)  LITER, P2new, P1new
      READ(5, *)  IX, IY, IZ
      read(5,*)  nrepl
      read(5 , 470)      PNAME
      read(5 , 470)      FNAME
470    format(A72) 

C      TRUE MODEL SPECIFICATION : 
C      ONLY NHK=2 IS SUPPORTED, NPI>= MAX0(NPLAG,NQLAG) 

      READ(5,*) NHK, NPI, THRE90, ND90
      DO 471  I =1 , NHK
      READ(5 , *) (XPHI(I , J), J =1 , NPI)
      READ(5 , *) (XTHETA(I , J), J =1 , NPI)       
471   CONTINUE

       close(unit  =  5)

      NBIT =  NBIT1 +  NBIT2
      NX  =  1 + NHP1 + NP + NQ
      NM  = NHP1 + MAX0(NPLAG , NQLAG)+1
      MCHRO=NBIT+(NOB-2*MSPAN)+NK*(NBIT1+NP*NBIT2+NBIT1+NQ*NBIT2)
      NV  =  1 +  NP + NQ

C       PI-WEIGHTS CALCULATION

      DO  472  II  =  1 , NHK
      DO  473   J =  1 , NPI
      PXHP(J) =  XPHI(II , J)
      PXHQ(J)  =  XTHETA(II , J)
473   CONTINUE
      CALL  PWEIG(PXHP, PXHQ, PXHPI, NPI, NPI, NPI)
      DO  474   J =  1 , NPI
      XPI(II , J)  = PXHPI(J) 
474   CONTINUE
472   CONTINUE

       open(unit= LPRINT, file=pname)

       write(LPRINT  ,  200)  
200    format(1h  ,  'XXGTARMA.FOR (mainpgm) Ottobre 2003'
     +/1h ,'GA estimate of k-regimes SETARMA subset lag model')
       write(lprint,  201) nrepl
201    format(1h   ,'number of replications nrepl = ', I10)
      WRITE(LPRINT, 204) NOB, NAC, MSPAN, NBIT,NBIT1,NBIT2, NK, ICRI
204   FORMAT(1H ,'NOB,NAC,MSPAN,NBIT,NBIT1,NBIT2,NK,ICRI= '
     +/1H    , I10, 7I5)
      write(LPRINT, 210) NL, P2, P1, MAXIT, XETA, IX, IY, IZ
210   format(1H ,'NL,P2,P1,MAXIT,XETA,IX,IY,IZ= '
     +/1H  ,I5,F7.2,F7.3,I10,D12.4,3I12)
      write(LPRINT, 211) NP, NQ, NPLAG, NQLAG, NX, MCHRO, NM, NDYP
211   format(1H ,'NP,NQ,NPLAG,NQLAG,NX,MCHRO,NM,NDYP= '/1H   , 8I5)
       write(LPRINT , 274)  
274    format(1h , 'data printout file')
       write(LPRINT , 275)      pname
       write(LPRINT , 284)  
284    format(1h , 'data file for storing distances')
       write(LPRINT , 275)      fname
275    format(1h  ,  A72)
      WRITE(LPRINT, 276) NHK, NPI, THRE90, ND90
276   FORMAT(1H  ,'TRUE MODEL : NHK,NPI,THRE90,ND90 ',2I5,F10.4,I5)
      DO  475  II =1  ,  NHK
      WRITE(LPRINT, 277) II
277   FORMAT(1H   ,'REGIME ',I5,5X,'LINE1:AR, LINE2:MA, LINE3:PI')
      WRITE(LPRINT, 278)  (XPHI(II , J), J =1 ,  NPI)
      WRITE(LPRINT, 278)  (XTHETA(II , J), J =1 ,  NPI)
      WRITE(LPRINT, 278)  (XPI(II , J), J =1 ,  NPI)
278   FORMAT(1H   ,  7F10.4)
475   CONTINUE
      WRITE(LPRINT, 309)
309   FORMAT(1H  ,'//')

C     MONTE CARLO EXPERIMENT RESULTS ARE SET TO ZERO AT THE START

      DO  310  II = 1 , NDYP
      DO  310  I = 1 , NK
      DO  310  JJ = 1 , NK
      DO 310   J  =  1, NP
310   WNP(II,I,JJ,J)  =  0.0
      DO  311  II  =1  ,  NDYP
      DO  311  I  =1  ,  NK
      DO 311   III =  1 ,  NP
      DO  311  JJ  =1  ,  NK
      DO  311  J =  1,  NPLAG
      WNPLAG(II,I,III,JJ,J)  =  0.0
      WNPP(II,I,III,JJ,J)  =  0.0
311   WNPP2(II,I,III,JJ,J)  =  0.0
      DO  313  II = 1 , NDYP
      DO  313  I = 1 , NK
      DO  313  JJ = 1 , NK
      DO 313   J  =  1, NQ
313   WNQ(II,I,JJ,J)  =  0.0
      DO  312  II  =  1  ,  NDYP
      DO  312  I  =  1  ,  NK
      DO 312   III =  1 ,  NQ
      DO  312  JJ  =  1  ,  NK
      DO  312  J =  1,  NQLAG
      WNQLAG(II,I,III,JJ,J)  =  0.0
      WNQQ(II,I,III,JJ,J)  =  0.0
312   WNQQ2(II,I,III,JJ,J)  =  0.0
      DO 314  II  =1  ,  NDYP
      DO 314  I  =1  ,  NK
      DO 314  JJ  =1  ,  NK
      WOBS(II,I,JJ)  =  0.0
      WOBS2(II,I,JJ) =  0.0
      WTHRE(II,I,JJ) = 0.0
314   WTHRE2(II,I,JJ) = 0.0
      DO  317   II  =  1 ,  NDYP
      DO  317   I  =  1 ,  NK
317   WCONTA(II,I)   =  0.0
      DO  316   II  =  1 ,  NDYP
      DO  316   I  =  1 ,  NK
316   WNK(II,I)   =  0.0
      DO  315  II =1 , NDYP
      DO  315  I =1 , NK
315   WNDYP(II,I)  =  0.0
      DO  318  II = 1,  NDYP
      DO  318  I =  1  , NK
      WVSSQ(II , I)  =  0.0 
318   WVSSQ2(II , I)  =  0.0 
      WVARE  =  0.0
      WVARE2  =  0.0
      WVAOB  =  0.0
      WVAOB2  =  0.0
      WAIC  =  0.0
      WAIC2  =   0.0
      WIT =   0.0

c        Monte Carlo loop
c        File for storing the distance values VSSQ

       open(unit= LDUMP, file=fname)

      do   50010  irepl = 1  , nrepl

       write(*,*) irepl

c        Generates time series normal unit white noise .

      vare  = 1D0
      nscart  =  500
      NTOT = NOB + NSCART
      NDR1 =  NOB - NM
      DO  401  I  =  1  ,  NTOT
402        PPX0=  uniform()
      PPX1 =  PPND16(PPX0 , IZAULT)
      IF(IZAULT .NE. 0)  GOTO  402
401   ETILDA(I) =  PPX1 * DSQRT(VARE)   

c        To change model, modify the following statements ...

       thre90 = 0D0
        nd90  =  1
       do  403  i =  1,  nm
403    z(i)  =  etilda(i)
        nm1  =  nm + 1
       do  404  i  = nm1 , ntot
       if(z(i - nd90) .le. thre90) then
       z(i) = 0.8 * z(i - 2) + etilda(i)
       else
       z(i) = -0.4 * z(i - 1) + etilda(i) - 0.4 * etilda(i - 2)
        endif
404    continue
       do  405  i =  1 ,  nob
405    z(i) = z(i +  nscart) 

c       open(unit=15,file='test90')
c       write(15,40510) (z(i), i=1,  nob)
c40510  format(F20.8)
c        close(unit=15)

c       Preliminary statistics computation

      CALL ZCOR(Z,NOB,NAC,RHO,COV,DM,DSD,IEXIT)                     
       dhm = 0D0
      ssql  =  0D0
      nm1 = nm + 1
      do  110  i = nm1 , nob
      dhm = dhm + z(i)
110   ssql = ssql + z(i) * z(i)
      dhm  =  dhm / dble(nob - nm)
	vhv  = ssql / dble(nob - nm) - dhm * dhm
      SSQL  =  vhv * dble(nob - nm)
       DO  33881 I  = 1  ,  NOB
33881  FZ(I)  =  Z(I)
        CALL GRADU(FZ,FITX,FITY,KKN,KNG,NOB)
       DO  33888 I  = 1  ,  NOB
33888  KKHN(I)  =  KNG(I)
       DO  33898 I  = 1  ,  NOB
33898  KKKN(I)  =  KKN(I)

C      WRITE(LPRINT,  187) IEXIT
C187   FORMAT(1H ,'Time series Z statistics ZCOR : IEXIT ', I5)
C      write(LPRINT, 207) dm, dsd, cov(1) 
C207   format(1h  ,'Input data mean,sqm,var =', 3D12.4)
C      write(6, 206) dhm ,ssql, vhv 
C206   format(1h  ,'Data nm+1,nob dhm,ssql,var =', 3D12.4)
C      write(LPRINT, 208) 
C208   format(1h ,'Autocorrelation function')
C       nac1  =  nac +  1
C      write(LPRINT, 203) (rho(j), j=1, nac1)
C203    format(1h  , 7f10.4)
c      write(lprint, 218) 
c218   format(1h  ,'time series values, ordered values, kkn, kkg')
c       do  72881  i  =1  ,  nob
c72881 write(lprint, 219) i, fz(i), fity(i), kkn(i), kng(i)
c219    format(1h   ,I10, 2F12.4, 2I10)
c      write(LPRINT, 608) 
c608   format(1h ,'Labels vector of observations in increasing ord KKHN')
c      write(LPRINT, 603) (kkhn(j), j=1, nob)
c603    format(1h  , 10I7)
C      write(lprint, 129) 
C129    format(1h    ,'//'  )

C      iprint =  1
C
C      FFFIT =  FFB(lchrom, mchro, z, mob, nob, nm, kkhn, kkkn, nd
C     +, r, MK, nk, k, nj, s2, nbit, mx, nx, NP, NQ, nxp, lahp, nxq, lahq  
C     +,icri,crki,xeta, X, RES, VFIT, XMAT, PHI, AUX1, VC1, BB, CC, IP,IQ
C     +, MSPAN, nchro, NBIT1, NBIT2,nplag,nqlag, NHP0, NHP1, NHP
C     +, iprint, lprint, ifault)
C
C      write(lprint, 1147)  ifault
C1147  format(1h   ,'Exit FFB, ifault = ', I5)

C    k-REGIMES THRESHOLD ARMA d, r1..rk, (p, l1..lp; q, l1..lq)-k times
C     THE GENETIC ALGORITHM FINDS ALL UNKNOWN PARAMETERS WITH BIC/AIC

      CALL  GTARMA(no, nn, nmx, nl, mchro, nchro, lchrom, lbest
     +, lchro1, lchro2, lchro3, lchro4, fbest, ssq, p2, p1, maxit, iter2
     +, z, mob, nob, nm, kkhn, kkkn, nd, r, MK, nk, k, nj, s2, nbit, mx
     +, nx, NP, NQ, nxp, lahp, nxq, lahq, icri, crki, xeta, X, RES, VFIT
     +, XMAT, PHI, AUX1, VC1, BB, CC, IP, IQ, MSPAN, naux1, naux2, fitnn
     +, NBIT1,NBIT2,LKYROM,NDYP,NPLAG,NQLAG,liter,p2new,p1new
     +, NHP0, NHP1, NHP, lprint, iexit)   

C      WRITE(LPRINT,  56187) IEXIT
C56187 FORMAT(1H ,'Exit genetic algorithm GTARMA : IEXIT ', I5)
C      write(lprint, 27771) (lkyrom(iiii), iiii=1,  mchro)
C27771 format(1h    , 70i1)

c      iprint =  1
      IPRINT  = 0

c      write(*,*) 'enter fffit, k = ', k

      FFFIT =  FFB(lkyrom, mchro, z, mob, nob, nm, kkhn, kkkn, nd
     +, r, MK, nk, k, nj, s2, nbit, mx, nx, NP, NQ, nxp, lahp, nxq, lahq  
     +,icri,crki,xeta, X, RES, VFIT, XMAT, PHI, AUX1, VC1, BB, CC, IP,IQ
     +, MSPAN, nchro, NBIT1, NBIT2,nplag,nqlag, NHP0, NHP1, NHP
     +,iprint, lprint, ifault)

c      write(*,*) 'exit fffit, k = ', k

C      write(lprint, 21147)  ifault
C21147  format(1h   ,'Exit FFB, ifault = ', I5)

c	do 260 i=1, nob
c	write(lprint, 21077)  z(i),vfit(i),res(i)
c21077 format(1h ,'z vfit res', 3F14.6)
c260	continue

C      WRITE(LPRINT , 300) 
C300   FORMAT(1H  ,'EXIT GTARMA : PARAMETER ESTIMATES ')
C      write(lprint, 113)  k
C113   format(1h   ,'No. of regimes k = ', I5)
C      WRITE(LPRINT,  308)  SSQL
C308   FORMAT(1H  ,'OBSERVATIONS SUM OF SQUARES SSQL = ', D14.6)
      OBSVAR = SSQL / DBLE(NOB - NM)
C      WRITE(LPRINT,  303)  OBSVAR
C303   FORMAT(1H  ,'OBSERVATIONS VARIANCE = ', D14.6)
        
       sumosq =  0D0
       nvalid  =  0
       AICCC =  0.0
	nHpara=0
        Do 367 I=1,k
       nHpara  =  nHpara + 1 +  nxp(i) +nxq(i)
       sumosq  =  sumosq  +  s2(i)
       nvalid =  nvalid + nj(i)
        AICCC = AICCC + DBLE(NJ(I)) * DLOG(S2(I) / DBLE(NJ(I)))
367     continue

c        SSQ  =  DOTPRD(Nob , Res , Res)
c      WRITE(LPRINT,  304)  SSQ
C      WRITE(LPRINT,  304)  sumosq
C304   FORMAT(1H  ,'RESIDUAL SUM OF SQUARES SSQ = ', D14.6)
c      resvar = SSQ / dble(nob - NM) 
c      WRITE(LPRINT,  305) resvar, NM 
        resvar  =  sumosq / dble(nvalid)
C      WRITE(LPRINT,  305) resvar, nvalid 
C305   FORMAT(1H  ,'RESIDUAL variance, Oss. eff. = ', D14.6, I5)
	
      AICCC =  (AICCC + 2D0 * DBLE(NHPARA)) / DBLE(NVALID)

C	WRITE(LPRINT,  3005) AICCC, nHpara 
C3005   FORMAT(1H  ,'AIC = ',4x, 'Numero parametri', D14.6, I5)
C      WRITE(LPRINT, 309)
 
C     STORING MONTE CARLO EXPERIMENT RESULTS 

c      WRITE(*,*) 'K ', K
c      WRITE(*,*) 'NHPARA ' , NHPARA
c      WRITE(*,*) 'TUTTI I PARAMETRI'
c      DO  3006  IIII =  1,  K
c      DO  3006  JJJJ =  1,  NHPARA
c3006  WRITE(*,*) JJJJ, PPP(IIII,JJJJ)
c      WRITE(*,*) '/'

      JNND  = ND
      JNNK  = K
      WCONTA(JNND,JNNK) =  WCONTA(JNND,JNNK) +  1
      DO   476  II  =  JNNK
      DO   476   J  =  1,  NPI  
      YPHI(II , J) =  0.0
      YTHETA(II , J) =  0.0
      YPI(II , J) =  0.0
476   CONTINUE
      DO  610  I = 1 , K
      JHHP =  NXP(I)
      IF(J.GT.0)  WNP(JNND,JNNK,I,JHHP)  =  WNP(JNND,JNNK,I,JHHP) + 1  
      JHHQ =  NXQ(I)
      IF(J.GT.0)  WNQ(JNND,JNNK,I,JHHQ)  =  WNQ(JNND,JNNK,I,JHHQ) + 1  
610   CONTINUE
      DO  611  I  =1  ,  K
      DO  477   JJJJ = 1  , NPI
      PXHP(JJJJ)   =  0.0
      PXHQ(JJJJ)   =  0.0
477   PXHPI(JJJJ)   =  0.0
      IPAHR =  0
      IPAHR  =  IPAHR  + 1
      JHHP =  NXP(I)
      IF(JHHP .GT.  0)  THEN
      DO  612  J =  1,  JHHP
      IJ =  LAHP(I,J)
      IF(IJ .GT. 0) THEN
      WNPLAG(JNND,JNNK,JHHP,I,IJ) = WNPLAG(JNND,JNNK,JHHP,I,IJ) + 1  
      IPAHR  =  IPAHR  + 1
c      WRITE(*,*) I, IPAHR, IJ, PPP(I,IPAHR)
      PXHP(IJ) = PPP(I,IPAHR) 
      WNPP(JNND,JNNK,JHHP,I,IJ)=WNPP(JNND,JNNK,JHHP,I,IJ)+PPP(I,IPAHR)  
      WNPP2(JNND,JNNK,JHHP,I,IJ)=WNPP2(JNND,JNNK,JHHP,I,IJ)
     ++PPP(I,IPAHR)**2
       ENDIF
612   CONTINUE
      ENDIF     
      JHHQ =  NXQ(I)
      IF(JHHQ .GT.  0)  THEN
      DO  613  J =  1,  JHHQ
      IJ =  LAHQ(I,J)
      IF(IJ .GT. 0) THEN
      WNQLAG(JNND,JNNK,JHHQ,I,IJ) = WNQLAG(JNND,JNNK,JHHQ,I,IJ) + 1  
      IPAHR  =  IPAHR  + 1
c      WRITE(*,*) I, IPAHR, IJ, PPP(I,IPAHR)
      PXHQ(IJ) = PPP(I,IPAHR) 
      WNQQ(JNND,JNNK,JHHQ,I,IJ)=WNQQ(JNND,JNNK,JHHQ,I,IJ)+PPP(I,IPAHR)  
      WNQQ2(JNND,JNNK,JHHQ,I,IJ) = WNQQ2(JNND,JNNK,JHHQ,I,IJ)
     + +  PPP(I,IPAHR) ** 2
       ENDIF
613   CONTINUE
      ENDIF
      CALL PWEIG(PXHP, PXHQ, PXHPI, NPI, NPI, NPI)
      DO  478  JJJJ =  1  , NPI
      YPHI(I , JJJJ)  =   PXHP(JJJJ)
      YTHETA(I , JJJJ)  =   PXHQ(JJJJ)
478   YPI(I , JJJJ)  =   PXHPI(JJJJ)
611   CONTINUE
      DO 614  I  =1  ,  K
      WOBS(JNND,JNNK,I)  =  WOBS(JNND,JNNK,I) + DBLE(NJ(I))
      WOBS2(JNND,JNNK,I)  =  WOBS2(JNND,JNNK,I) + DBLE(NJ(I)) ** 2 
      IF(I .LT. K) THEN 
      WTHRE(JNND,JNNK,I)  =  WTHRE(JNND,JNNK,I) + R(I)
      WTHRE2(JNND,JNNK,I)  =  WTHRE2(JNND,JNNK,I) + R(I) ** 2
       ENDIF
614   CONTINUE
      WNDYP(JNND,JNNK)  =  WNDYP(JNND,JNNK) +  1
      WNK(JNND,JNNK)  =  WNK(JNND,JNNK) +  1
c      WRITE(LPRINT, 279) JNNK, NPI, R(1), ND90
c279   FORMAT(1H  ,'ESTIM MODEL : JNNK,NPI,R(1),ND90 ',2I5,F10.4,I5)
c      DO  479  II =1  ,  JNNK
c      WRITE(LPRINT, 280) II
c280   FORMAT(1H   ,'REGIME ',I5,5X,'LINE1:AR, LINE2:MA, LINE3:PI')
c      WRITE(LPRINT, 281)  (YPHI(II , J), J =1 ,  NPI)
c      WRITE(LPRINT, 281)  (YTHETA(II , J), J =1 ,  NPI)
c      WRITE(LPRINT, 281)  (YPI(II , J), J =1 ,  NPI)
c281   FORMAT(1H   ,  7F10.4)
c479   CONTINUE
      VSSQ  =  0.0
      DO   480  IIII = 1 ,  JNNK
      DO   480   JJJJ  =  1 , NPI
480   VSSQ = VSSQ + (XPI(IIII,JJJJ)-YPI(IIII,JJJJ)) ** 2
      WVSSQ(JNND,JNNK)  =  WVSSQ(JNND,JNNK) +  VSSQ
      WVSSQ2(JNND,JNNK)  =  WVSSQ2(JNND,JNNK) +  VSSQ ** 2
      WVAOB  =  WVAOB + OBSVAR
      WVAOB2  =  WVAOB2 +  OBSVAR ** 2
      WVARE  =  WVARE +  RESVAR
      WVARE2  =  WVARE2 +	RESVAR ** 2
      WAIC  =  WAIC +	 AICCC 
      WAIC2  =   WAIC2 +  AICCC ** 2 
      WIT =   WIT +  ITER2

c      PAUSE

       WRITE(LDUMP, 1716) VSSQ
1716   FORMAT(F12.6)

50010   continue

C     MONTE CARLO RESULTS

      DO  701  II  =  1 ,  NDYP
      DO  701   JJ =  1  ,  NK
      JNND  =  II
      JNNK =  JJ
      WCO = WCONTA(JNND,JNNK)
      WRITE(LPRINT, 1701) JNND , JNNK, WCO
1701  FORMAT(1H   / 1H   ,'D  ', I5,  5X, 'K ', I5, 5X, 'occurs ',5F5.0)
      IF(WCO .GT. 0D0) THEN
      WVSQ  =  WVSSQ(JNND,JNNK) / WCO
      WVSQ2 = DSQRT(DABS(WVSSQ2(JNND,JNNK) / WCO - WVSQ ** 2)) 
      WRITE(LPRINT, 1713) WVSQ
1713  FORMAT(1H  ,'average pi-weights mean square error', F12.4)
      WRITE(LPRINT, 1715) WVSQ2
1715  FORMAT(1H  ,'stand dev pi-weights mean square error', F12.4)
      DO  704  I  = 1 ,  JNNK
      WO  =  WOBS(JNND,JNNK,I) / WCO
      WO2 = DSQRT(DABS(WOBS2(JNND,JNNK,I) / WCO - WO ** 2)) 
      WTH  =  WTHRE(JNND,JNNK,I) / WCO
      WTH2 = DSQRT(DABS(WTHRE2(JNND,JNNK,I) / WCO - WTH ** 2)) 
      WRITE(LPRINT,1706) I, WO, WO2, WTH, WTH2 
1706  FORMAT(1H  ,'regime,n(mean,se)', I5, 5X, 2F10.0 
     +/1H  , 'threshold(mean,se)', 5X, 2F12.4)
      DO  702  J  =  1 , NP
      JHHP  =    J
      IF(WNP(JNND,JNNK,I,JHHP) .GT. 0D0) THEN
      WRITE(LPRINT, 1702) JHHP ,  WNP(JNND,JNNK,I,JHHP)  
1702  FORMAT(1H   ,'p = ', I5, 5X, 'times for regime ', 5F5.0)
C      WRITE(LPRINT, 1703) 
C1703  FORMAT(1H ,'reg','lag',5X
C     +,'occurs',5X,' parameter',5X,'  st. dev.')
      DO  703  ILAG =1 , NPLAG
      WN = WNPLAG(JNND,JNNK,JHHP,I,ILAG)
      IF(WN .GT. 0D0)  THEN
      WNPPN = WNPP(JNND,JNNK,JHHP,I,ILAG)/WN
      WNPPN2 = DSQRT(DABS(WNPP2(JNND,JNNK,JHHP,I,ILAG)/WN - WNPPN ** 2)) 
C      WRITE(LPRINT, 1704)  I, ILAG , WN , WNPPN , WNPPN2  
C1704  FORMAT(1H  ,2I3, 5X, F6.0 , 5X , F10.4 , 5X , F10.2)
      WRITE(LPRINT, 1704)  ILAG , WN , WNPPN , WNPPN2  
1704  FORMAT(1H ,'lag ', I5, 'occurs ', F5.0, 'mean,se ', 2F12.4)
       ENDIF
703   CONTINUE
       ENDIF
702   CONTINUE
      DO  705  J  =  1 , NQ
      JHHQ  =    J
      IF(WNQ(JNND,JNNK,I,JHHQ) .GT. 0D0) THEN
      WRITE(LPRINT, 1705) JHHQ ,  WNQ(JNND,JNNK,I,JHHQ)  
1705  FORMAT(1H   ,'q = ', I5, 5X, 'times for regime ', 5F5.0)
C      WRITE(LPRINT, 1708) 
C1708  FORMAT(1H ,'reg','lag',5X
C     +,'occurs',5X,' parameter',5X,'  st. dev.')
      DO  706  ILAG =1 , NQLAG
      WN = WNQLAG(JNND,JNNK,JHHQ,I,ILAG)
      IF(WN .GT. 0D0)   THEN
      WNQQN = WNQQ(JNND,JNNK,JHHQ,I,ILAG) / WN
      WNQQN2 = DSQRT(DABS(WNQQ2(JNND,JNNK,JHHQ,I,ILAG)/WN - WNQQN ** 2)) 
C      WRITE(LPRINT, 1712)  I, ILAG , WN , WNQQN , WNQQN2  
C1712  FORMAT(1H  ,2I3, 5X, F6.0 , 5X , F10.4 , 5X , F10.2)
      WRITE(LPRINT, 1712)  ILAG , WN , WNQQN , WNQQN2  
1712  FORMAT(1H ,'lag ', I5, 'occurs ', F5.0, 'mean,se ', 2F12.4)
       ENDIF
706   CONTINUE
       ENDIF
705   CONTINUE
704   CONTINUE
       ENDIF
      WRITE(LPRINT, 1707)
1707  FORMAT(1H  ,'/')
701   CONTINUE
      WVAOB  =  WVAOB / DBLE(NREPL)
      WVAOB2  =  DSQRT(DABS(WVAOB2 / DBLE(NREPL) - WVAOB ** 2))
      WRITE(LPRINT, 1714) WVAOB, WVAOB2
1714  FORMAT(1H ,'observation variance (mean,se)', 2F14.4)      
      WVARE  =  WVARE / DBLE(NREPL)
      WVARE2  =  DSQRT(DABS(WVARE2 / DBLE(NREPL) - WVARE ** 2))
      WRITE(LPRINT, 1709) WVARE, WVARE2
1709  FORMAT(1H ,'residual variance (mean,se)', 2F14.4)      
      WAIC  =  WAIC / DBLE(NREPL) 
      WAIC2  =   DSQRT(DABS(WAIC2 / DBLE(NREPL) - WAIC ** 2)) 
      WRITE(LPRINT, 1710) WAIC, WAIC2
1710  FORMAT(1H ,'             NAIC (mean,se)', 2F14.4)      
      WIT =   WIT / DBLE(NREPL)
      WRITE(LPRINT, 1711) WIT
1711  FORMAT(1H ,'iterations needed (average)', F14.0)      

      CLOSE(UNIT=LDUMP)
      CLOSE(UNIT=LPRINT)
      STOP
      END


      SUBROUTINE  GTARMA(no, nn, nmx, nl, mchro, nchro, lchrom, lbest
     +, lchro1, lchro2, lchro3, lchro4, fbest, ssq, p2, p1, maxit, iter2
     +, z, mob, nob, nm, kkhn, kkkn, nd, r, MK, nk, k, nj, s2, nbit, mx
     +, nx, NP, NQ, nxp, lahp, nxq, lahq, icri, crki, xeta, X, RES, VFIT
     +, XMAT, PHI, AUX1, VC1, BB, CC, IP,IQ, MSPAN, naux1, naux2, fitnn
     +, NBIT1,NBIT2,LKYROM,NDYP,NPLAG,NQLAG,liter,p2new,p1new
     +, NHP0, NHP1, NHP, lprint, iexit)   

      IMPLICIT REAL*8(A - H , O - Z)

C     TIME SERIES DIMENSION

      DIMENSION Z(NOB), X(NOB), KKHN(NOB), XMAT(MOB, NX), PHI(NX)
     +, AUX1(NX), VC1(MX, NX), BB(NX), CC(NX), IP(NX), IQ(NX)
     +, RES(NOB), VFIT(NOB)
     +, FZ(NOB), FITX(NOB), FITY(NOB), KKN(NOB), KNG(NOB), KKKN(NOB)
     +, NHP(NK), nxp(NK), LAHP(MK, NP), nxq(MK), LAHQ(MK, NQ) 

C     THRESHOLD PARAMETERS DIMENSION

      DIMENSION R(NK), NJ(NK), S2(NK)

C     GENETIC ALGORITHM DIMENSION

      dimension  NO(nmx, Mchro), NN(nmx, Mchro), lbest(Mchro)
     +,lchrom(Mchro), lchro1(Mchro), lchro2(Mchro)
     +, lchro3(Mchro), lchro4(Mchro), LKYROM(MCHRO)
     +, naux1(nbit), naux2(nx), FITCUM(nl), FITN(nl), fitnn(nl)

      data  zero /0.0D-000/
	data small , big / 8.43D-37, 3.37D+38/
        data  tol  /8.43D-37/ 

        CRKIQ  =  BIG
        do  47127  i  =1  ,  mchro
47127         LKYROM(I)  =  0

       DO  47470  IDYP  =  1 ,  NDYP

C       Initializations

       DO 99399   I   =   1  ,  NL
       DO 99399    J   =    1  ,   MCHRO
       NO(I   ,   J)   =  0
99399     NN(I  ,   J)  =   0
        DO 99398  I  =   1  ,  MCHRO
99398   LBEST(I)   =   0
        DO 99397  I  =   1  ,  MCHRO
        lchro1(I)   =   0
          lchro2(I)   =   0
        lchro3(I)   =   0
99397     lchro4(I)   =   0
       DO 99396  I  =  1,  NL
        FITCUM(I)= zero
        FITN(I)  =   zero
99396    FITNN(I)  =   zero
        DO 99395  I  =   1  , NBIT
99395   NAUX1(I)   =   0
        DO 99394  I  =   1  , NX
99394   NAUX2(I)   =   0
        iprint  =  0

C       Initial population
C       The genes for threshold search range from mspan+1 to nob-mspan
C       No more than nk-1 one's are allowed

       do  33440 kk = 1 , nl

c   repeat if generation procedure fails to produce an admissible result

96995 continue
       do  145 i =  1 , mchro
145    lchrom(i)  =  0

c       delay parameter generation: it must be greater than zero     
C        ITS VALUE IS IDYP AND I UPDATED BY THE LOOP 47470
C96994  continue
C       iczero =  0
C       do  33441 i =1 , nbit
C       lchrom(i)  =  idnint(random(l))
C       if(lchrom(i) .gt. 0) iczero = 1
C33441  continue
C       if(iczero  .le.  0) goto  96994

      CALL INTOBI(naux1, nbit, IDYP)     
      DO  47441  I  =1  ,  NBIT
47441  LCHROM(I)  =  NAUX1(I)

c       number of threshold parameters k-1  and thresholds  r(1)..r(k-1)
c       the number of regimes is k 

       nk1  =  nk - 1
       ris = random(l)
      nk2 =  max0(0, min0(nk1, idnint(-5D-1 + ris * (dble(nk1) + 1D0)))) 

       n1  =  mspan + 1
       n2 =  nob - mspan
       i = nbit
       DO 96997  ii = n1, n2
       i=  i + 1
96997  lchrom(i) = 0

c       nk2 is the number  of threshold parameters 
c       if  nk2<=0 there is nothing to locate

       if(nk2  .le. 0)  goto 96996
       do  33443  j  =  1  ,  nk2
       ris =  random(l)
       ij = max0(n1, min0(n2, idnint(dble(n1)+ris*(dble(n2)-dble(n1)))))
       ij = ij - mspan + nbit
33443  lchrom(ij) = 1
96996  continue

c       for each regime, generates the subset ARMA structure
c       regimes are nk3 = nk2 + 1

       nk3  =  nk2 + 1

       DO 99997  II  =   1  ,  nk3

c       subset autoregressive structure
   
       ncp  =  0
       nlagh  =  0
       do  34445 ij = 1 , np
       do 35555  jj = 1 , nbit2
35555  naux1(jj) =  idnint(random(l)) 
       call ndec1(naux1, nbit2, llagh) 
       if(llagh  .le. 0)  goto  35556
       ncp =  ncp  + 1
       nlagh =  nlagh + llagh
       if(ncp .gt. np. or. nlagh .gt. npLAG)  then
       ncp =  ncp - 1
       goto  35556
       else
       naux2(ncp) = nlagh
       endif
34445  continue
35556  continue
       call  intobi(naux1, nbit1, ncp)
       do  35557  jj = 1 , nbit1
       i = i + 1
35557  lchrom(i)  =  naux1(jj)
       if(ncp. gt .0)  then
       do   35558  ij = 1 ,  ncp
       llagh =  naux2(ij)
       call  intobi(naux1, nbit2, llagh)
       do  35559  jj = 1 , nbit2
       i = i + 1
35559  lchrom(i)  =  naux1(jj)
35558  continue             
       endif

c       subset moving average structure
   
       ncq  =  0
       nlagh  =  0
       do  24445 ij = 1 , nq
       do 25555  jj = 1 , nbit1
25555  naux1(jj) =  idnint(random(l)) 
       call ndec1(naux1, nbit2, llagh) 
       if(llagh  .le. 0)  goto  25556
       ncq =  ncq  + 1
       nlagh =  nlagh + llagh
       if(ncq .gt. nq. or. nlagh .gt. nqLAG)  then
       ncq =  ncq - 1
       goto  25556
       else
       naux2(ncq) = nlagh
       endif
24445  continue
25556  continue
       call  intobi(naux1, nbit1, ncq)
       do  25557  jj = 1 , nbit1
       i = i + 1
25557  lchrom(i)  =  naux1(jj)
       if(ncq. gt .0)  then
       do   25558  ij = 1 ,  ncq
       llagh =  naux2(ij)
       call  intobi(naux1, nbit2, llagh)
       do  25559  jj = 1 , nbit2
       i = i + 1
25559  lchrom(i)  =  naux1(jj)
25558  continue             
       endif

99997  continue

       nchro = i

c      fill with zeroes the last bits of the current string

       nchro1 =  nchro + 1
       if(nchro1 .le. mchro) then
       do  33449 i =  nchro1 , mchro
33449  lchrom(i)  =  0
       endif

c      The routine checks the chromosome and computes its fitness
c      If checking pass, then stores the chromosome,
c      otherwise generates another

      FFFIT =  FFB(lchrom, mchro, z, mob, nob, nm, kkhn, kkkn, nd
     +, r, MK, nk, k, nj, s2, nbit, mx, nx, NP, NQ, nxp, lahp, nxq, lahq  
     +,icri,crki,xeta, X, RES, VFIT, XMAT, PHI, AUX1, VC1, BB, CC, IP,IQ
     +, MSPAN, nchro, NBIT1,NBIT2,nplag,nqlag
     +, NHP0, NHP1, NHP,iprint, lprint, ifault)

c      write(lprint, 1147)  kk, ifault
c1147  format(1h   ,'Ind. ', I6, 4X, 'Exit FFB, ifault = ', I5)

       IF(IFAULT .NE. 0) goto 96995

c      The generated chromosome is admissible, store it and continue

       do  33448  i = 1 , mchro
33448  no(kk , i) =  lchrom(i) 
       fitn(kk) = fffit

33440  continue

C       Genetic algorithm loop

         iter  =  0
         iter2    =  0
         lhiter = 0
	   fbest1  =   0.0

c	 WRITE(*,*)  '     Iterazione ..... '

1234	  continue

         iter  =  iter + 1

c	WRITE(*,*)  ITER

C   Fitness function computation for each individual
C   The largest and smallest fitness are computed and inviduals recorded

	FBEST  =  zero
      fwost  =   big
	DO 2000 KK = 1 , NL
      fffit = fitn(kk)
	if (fffit.le.fbest) goto 2043
	fbest  =   fffit
	nbest  =    kk
2043     continue
	if (fffit.ge.fwost) goto 2044
	fwost  =   fffit
	nwost  =    kk
2044     continue
2000    CONTINUE

	  iter1  =  iter  -   1
c      if(fbest-fbest1 .gt. tol) write(LPRINT,225) iter1, fbest1 , fbest
c225     format(1h ,4x,i10,' : ',D12.5,' -> ',D12.5)
       if(fbest-fbest1 .gt. tol) then
       iter2 =  iter1
       lhiter  =  0
       else
       lhiter =  lhiter + 1
       endif
      if(lhiter  .ge. liter) then
       p2 = p2new
       p1  =  p1new
      endif

C        Elitist strategy

      if(fbest1 .le. fbest)  then
       do  2002  i =1  ,  mchro
2002  lbest(i)  =  no(nbest , i)
       fbest1  =  fbest
      else
       do  2003  i =1  , mchro
2003  no(nwost , i)  =  lbest(i)
       fitn(nwost) =  fbest1
       nbest =  nwost
      fbest  =  fbest1
      endif

c      WRITE(LPRINT,  55593)  
c55593  FORMAT(1H  ,'Elitist strategy')
c      write(lprint, 55494)
c55494 format(1h   ,'best indiv., fitness')
c	 WRITE(LPRINT,55500) NBEST, FBEST
c55500  format(1h  ,I10,  E15.6) 
c       write(LPRINT , 55561)  (LBEST(I), I=1,mchro) 
c55561	 format(1h   ,70I1)
c      write(lprint, 55484)
c55484 format(1h   ,'worst indiv., fitness')
c	 WRITE(LPRINT,57500) NWOST, FWOST
c57500  format(1h  ,I10,  E15.6) 
c      WRITE(LPRINT,  55592)
c55592  format(1h   ,'Population ')
c       do 76011  kk  =  1  ,   nl
c76011  write(lprint, 54557) (no(kk ,  j),  j=1, mchro)
54557   format(1h   ,70I1)

	  IF(ITER.gt.MAXIT) GOTO  2100

C   Selection to form the new generation: P(t) ---> P*(t+1)
C   Complete replacement (generation gap G=1) and Roulette wheel rule

	FFTOT = zero
	DO 22000 KK = 1 , NL
       FFFIT =  FITN(KK)
	FFTOT = FFTOT + FFFIT
	 FITCUM(KK)  =   FFTOT
22000    CONTINUE

          if(fftot .LT. SMALL)    then
          iexit  =  -1
          return
          endif

	  DO  22007  KK =  1 ,  NL
22007  FITCUM(KK)  =   FITCUM(KK)/FFTOT
	  
	DO 3000 KK = 1 , NL
       ris = random(l)
       DO 3001  KS  =  1  ,   NL
       IF(RIS.LE.FITCUM(KS))  GOTO 3003
3001    CONTINUE
	 KS  =  NL
3003    fitnn(kk)  =  fitn(ks)
        DO 3100 I = 1 , mchro
3100    NN(kk , i) = NO(ks , i)
3000    CONTINUE

	 DO 3010   KK  =  1  ,  NL
       fitn(kk)   =  fitnn(kk)
	 DO 3010    I  =  1  ,  mchro
3010     NO(kk , i)  =  NN(kk , i)

c      WRITE(LPRINT,  55594)  
c55594  FORMAT(1H  ,'Selection ')
c       do 76012  kk  =  1  ,   nl
c76012  write(lprint, 54557) (no(kk , j),  j=1, mchro)

C      Uniform crossover is performed, slightly modified to take the 
C      chromosome fragments meaning into account: P*(t+1) ---> P**(t+1) 
C   ...  Uniform crossover has been replaced by exchange of delay
C        parameter and a single exchang of threshold parameter
C      nl * p2 / 2 pairs are chosen at random, each individual may be 
C      enter a pair more than once

      nbit11  =  nbit +  1
      nxit = nbit + (nob - 2 * mspan)
      ncrott  =  idint(p2 * dble(nl) / 2D0)

      DO     30     jcrott  =   1    ,   ncrott

        ris = random(l)
       kk1  =  MAX0(1, MIN0(NL, IDnint(5D-1   +   ris * DBLE(nl)))) 
        ris = random(l)
       kk2 = MOD(IDNINT(5D-1 + ris * DBLE(nl-1)) + kk1 -1, nl) + 1 

C       write(lprint,  13117) jcrott
C13117  format(1h  , 'coppia ',  i5)
C      write(lprint,  13118) kk1 , kk2
C13118 format(1h  ,'kk1,kk2 ', 2I6)

C Uniform crossover: a binary mask is generated which tells if exchange
C between bits has to be made (1=yes, 0=no). This applies to the first
C nbit genes (the delay parameter) and the threshold flags (genes from 
C nbit+1 to nbit+(nob-2*mspan)
C   ...  Uniform crossover has been replaced by
C        (1) exchange delay parameter
C        (2) transfer threshold parameter and/or exchange subset ARMA
C            models in a single location

c      do 31881 i =1 , mchro
c31881 lchrom(i)  =   0
c      do  31  i  =1  ,  nxit
c31    lchrom(i)   =  idnint(random(l))
c
c      write(lprint, 54880) 
c54880 format(1h   ,'mask vector ')
c      write(lprint, 54557) (lchrom(i), i =1 , mchro)

      do 61881 i =1 , mchro
61881 lchrom(i)  =   0
      do  61882  i  =nbit11  ,  nxit
61882 lchrom(i)   =  no(kk1, i) +  no(kk2, i)
      iki  =  0
      do  61883  i  =  nbit11 , nxit
       if(lchrom(i)  .gt.  0) then
       iki  =  iki  + 1
       kkn(iki)  =  i
       endif
61883 continue

C      write(lprint, 64880)
C64880 format(1h    ,'sum vector lchrom ')
C      write(lprint, 54557) (lchrom(i), i =1 , mchro)
C      write(lprint, 64990)  iki
C64990 format(1h   ,'number of nonzero locations iki = ', I5)
C      write(lprint, 64991) (kkn(i),  i=1,  iki)
C64991 format(1h   ,14I5)      

      do  3004  i =1  , mchro
      lchro1(i)  =  no(kk1 , i)
      lchro3(i) =  no(kk1 , i)
      lchro2(i)  =  no(kk2 , i)
3004  lchro4(i) =  no(kk2 , i)

C      write(lprint, 54882) 
C54882 format(1h   ,'parents 1-2. offsprings 1-2')
C      write(lprint, 54557) (no(kk1 , jjjj),  jjjj=1, mchro)
C      write(lprint, 54557) (no(kk2 , jjjj),  jjjj=1, mchro)

      ichang  =  0

C       Uniform crossover on the delay parameter
C
C      do  33  i  =1  ,  nbit
C      if(lchrom(i)  .le. 0)  then
C      lchro3(i)  =  lchro1(i)
C      lchro4(i)  =  lchro2(i)
C      else
C      lchro3(i)   =  lchro2(i)
C      lchro4(i)  = lchro1(i)
C      if(lchro1(i) .ne. lchro2(i)) ichang = 1
C      endif
C33    continue

C        The delay parameter exchanges between chromosomes

C      do  61033   i  =1  ,  nbit
C       if(lchro1(i) .ne. lchro2(i)) then
C       ichang  =  1
C       lchro3(i)   =  lchro2(i)
C       lchro4(i)  = lchro1(i)
C       endif
C61033 continue

C       In this version all delay parameter values are tried
C       The current delay parameter value is to  be left unchanged 

C       Uniform crossover on the thresholds
C   ...  Replaced by crossover on a single location

      nxit1  =  nxit
      nxit2  =  nxit
      nxit3 = nxit
      nxit4  = nxit

C       The chromosome fragment after the first nxit bits changes
C       according to the bit exchange in the threshold part
C       No exchange takes place if there is only a single regime in both
C       chromosomes

      if(iki  .gt.  0)  then

      ltag = max0(1, min0(iki, idnint(1D0 + random(l) * dble(iki - 1))))
      lbit =  kkn(ltag)

C      write(lprint, 64992) ltag , lbit
C64992 format(1h  ,'exchange regime ltag = ', I5
C     +/1h    ,'threshold location lbit = ', I5)
 
C      do  34  i  = nbit1,  nxit

       i  =  lbit

      if(lchrom(i)  .le.  0)  then
      lchro3(i) =  lchro1(i)
       if(lchro1(i) .ge.  1)  then
       call ksarma(naux1,nbit,NBIT1,NBIT2,lchro1,mchro,nxit1,klarma)
       nxit0 =  nxit1 + 1
       nxit1 =  nxit1 + klarma
       do  37  iki =  nxit0, nxit1
       nxit3 = nxit3 + 1
37     lchro3(nxit3)  =  lchro1(iki)
       endif       
      lchro4(i) =  lchro2(i)
       if(lchro2(i) .ge.  1)  then
       call ksarma(naux1,nbit,NBIT1,NBIT2,lchro2,mchro,nxit2,klarma)
       nxit0 =  nxit2 + 1
       nxit2 =  nxit2 + klarma
       do  38  iki =  nxit0, nxit2
       nxit4 = nxit4 + 1
38     lchro4(nxit4)  =  lchro2(iki)
       endif       
      else
      if(lchro3(i) .ne. lchro2(i)) ichang  =  1
      lchro3(i) =  lchro2(i)
       if(lchro2(i) .ge.  1)  then
       call ksarma(naux1,nbit,NBIT1,NBIT2,lchro2,mchro,nxit2,klarma)
       nxit0 =  nxit2 + 1
       nxit2 =  nxit2 + klarma
       do  39  iki =  nxit0, nxit2
       nxit3 = nxit3 + 1
       if(lchro3(nxit3) .ne. lchro2(iki)) ichang  =1 
39     lchro3(nxit3)  =  lchro2(iki)
       endif       
      if(lchro4(i) .ne. lchro1(i)) ichang  =  1
      lchro4(i) =  lchro1(i)
       if(lchro1(i) .ge.  1)  then
       call ksarma(naux1,nbit,NBIT1,NBIT2,lchro1,mchro,nxit1,klarma)
       nxit0 =  nxit1 + 1
       nxit1 =  nxit1 + klarma
       do  40  iki =  nxit0, nxit1
       nxit4 = nxit4 + 1
       if(lchro4(nxit4) .ne. lchro1(iki)) ichang  =1 
40     lchro4(nxit4)  =  lchro1(iki)
       endif
      endif

c34    continue

      endif

C     The last regime exchanges between chromosomes with probability 1/2
C   ...  Under new crossover, the last regime never exchanges

C       ixrr = idnint(random(l)) 
C
C      write(lprint, 56781) ixrr
C56781 format(1h   ,'Last regime switch ixrr ', I5)

         IXRR  =  0

       if(ixrr .le. 0) then
       call ksarma(naux1,nbit,NBIT1,NBIT2,lchro1,mchro,nxit1,klarma)
       nxit0 =  nxit1 + 1
       nxit1 =  nxit1 + klarma
       do  47  iki =  nxit0, nxit1
       nxit3 = nxit3 + 1
47     lchro3(nxit3)  =  lchro1(iki)
       call ksarma(naux1,nbit,NBIT1,NBIT2,lchro2,mchro,nxit2,klarma)
       nxit0 =  nxit2 + 1
       nxit2 =  nxit2 + klarma
       do  48  iki =  nxit0, nxit2
       nxit4 = nxit4 + 1
48     lchro4(nxit4)  =  lchro2(iki)
       else
       call ksarma(naux1,nbit,NBIT1,NBIT2,lchro2,mchro,nxit2,klarma)
       nxit0 =  nxit2 + 1
       nxit2 =  nxit2 + klarma
       do  49  iki =  nxit0, nxit2
       nxit3 = nxit3 + 1
       if(lchro3(nxit3) .ne. lchro2(iki)) ichang  =1 
49     lchro3(nxit3)  =  lchro2(iki)
       call ksarma(naux1,nbit,NBIT1,NBIT2,lchro1,mchro,nxit1,klarma)
       nxit0 =  nxit1 + 1
       nxit1 =  nxit1 + klarma
       do  50  iki =  nxit0, nxit1
       nxit4 = nxit4 + 1
       if(lchro4(nxit4) .ne. lchro1(iki)) ichang  =1 
50     lchro4(nxit4)  =  lchro1(iki)
       endif

C       Fill with zeroes the rest of chromosomes lchro3 and lchro4

      nxit31 = nxit3 + 1
      if(nxit31 .le. mchro) then
      do  3030  iki =  nxit31 , mchro
3030  lchro3(iki)  =  0
      endif
      nxit41 = nxit4 + 1
      if(nxit41 .le. mchro) then
      do  3031  iki =  nxit41 , mchro
3031  lchro4(iki)  =  0
      endif

C    Checks if crossover produced really new offsprings
C    Unless ichang is different from 0, the chromosomes remain unchanged

        if(ichang .eq.  0)  goto 430

C       Checks the offsprings lchro3 and lchro4 if admissible
C       Crossover takes place only if BOTH offsprings are legal

      CALL  DECODE(LCHRO3, MCHRO, NCHRO, Z, NM, NOB, KKHN, ND, R, MK, NK
     +, NP, NQ, NXP, LAHP, NXQ, LAHQ,NBIT,NBIT1,NBIT2,MSPAN,K,K1
     +,nplag,nqlag,ixfla3)
      CALL  DECODE(LCHRO4, MCHRO, NCHRO, Z, NM, NOB, KKHN, ND, R, MK, NK
     +, NP, NQ, NXP,LAHP,NXQ,LAHQ,NBIT,NBIT1,NBIT2,MSPAN, K, K1
     +, nplag, nqlag, ixfla4)

       if(ixfla3.eq.0  .and.  ixfla4.eq.0) then
      FFFIT1 =  FFB(lchro3, mchro, z, mob, nob, nm, kkhn, kkkn, nd
     +, r, MK, nk, k, nj, s2, nbit, mx, nx, NP, NQ, nxp, lahp, nxq, lahq  
     +,icri,crki,xeta, X, RES, VFIT, XMAT, PHI, AUX1, VC1, BB, CC, IP,IQ
     +, MSPAN, nchro, NBIT1,NBIT2, nplag, nqlag
     +, NHP0, NHP1, NHP, iprint, lprint, ixfla1)
      FFFIT2 =  FFB(lchro4, mchro, z, mob, nob, nm, kkhn, kkkn, nd
     +, r, MK, nk, k, nj, s2, nbit, mx, nx, NP, NQ, nxp, lahp, nxq, lahq  
     +,icri,crki,xeta, X, RES, VFIT, XMAT, PHI, AUX1, VC1, BB, CC, IP,IQ
     +, MSPAN, nchro, NBIT1,NBIT2, nplag, nqlag
     +, NHP0, NHP1, NHP, iprint, lprint, ixfla2)
       else
      goto   430
       endif

       if(ixfla1.eq.0  .and.  ixfla2.eq.0) then
      do  3005  i =  1  ,  mchro
      no(kk1 , i)  =   lchro3(i)
3005  no(kk2 , i)  =  lchro4(i)
      fitn(kk1)  = fffit1
      fitn(kk2)  = fffit2
       endif

430   continue

C      write(lprint,  13119)  ixfla1, ixfla2, ixfla3, ixfla4
C13119 format(1h  ,'Error flags 4 chromosomes ', 4I4)
C      write(lprint, 54557) (no(kk1 , jjjj),  jjjj=1, mchro)
C      write(lprint, 54557) (no(kk2 , jjjj),  jjjj=1, mchro)

30        continue
	  
C      WRITE(LPRINT,  55595)  
C55595  FORMAT(1H  ,'Crossover ')
C       do 76022  kk  =  1  ,   nl
C76022  write(lprint, 54557) (no(kk , j),  j=1, mchro)

C        Mutation: P**(t+1) ---> P***(t+1)
C        For any given individual, each bit flips if RND <= Pm  (Pm=P1)
C        The mutate chromosome is adjusted for legalization if needed

      nxit = nbit + (nob - 2 * mspan)

	DO  4000   KK  =  1 ,  NL
      ichang  =  0
      do  4004  i =1  , mchro
4004   lchrom(i) =  no(kk , i)

C       write(lprint, 40513)  kk
C40513  format(1h   /1h     ,'Mutation: individual n. ', I5)
C       write(lprint, 54557) (lchrom(i),  i=1, mchro)

C       Mutation on the delay parameter

C      write(lprint, 40515)
C40515 format(1h   , 'mutation on delay starts' )
C      WRITE(lprint, 54557) (LCHROM(JJJJ), JJJJ=1 , MCHRO) 

C      do  4221  i  = 1 ,  nbit
C4221  naux1(i)  =  lchrom(i)
C      call ndec1(naux1, nbit, ndd)

c      write(lprint, 40511) ndd
c40511	format(1h   ,'initial delay parameter ', I5)

C      jchang = 0
C	DO   4200   I  =  1  ,  nbit
C	IF(random(l).gt.P1) goto  4200
C      jchang = 1
      
c      write(lprint, 40517) i, jchang
c40517 format(1h   ,'i = ', I5, 5X, 'jchang = ', I5)

C      naux1(i) =  1  - naux1(i) 
C4200    CONTINUE
C      call ndec1(naux1, nbit, ndd)
C      if(ndd  .le.  0)  then
C      jchang  =  0

c      write(lprint, 40512)
c40512 format(1h  ,'delay parameter mutates to zero')

C      else
C       if(jchang  .eq.  1)  then
C       ichang  =  1
C       do  4222  i  =1  ,  nbit
C4222   lchrom(i)  =   naux1(i)

c       write(lprint, 40514) ndd
c40514	 format(1h   ,'mutated delay parameter ', I5)

C       else

c       write(lprint, 40516)
c40516  format(1h  ,'delay parameter is left unchanged')

C       endif
C      endif

C      write(lprint, 40518) ichang
C40518 format(1h  ,'end delay par. mutation, ichang = ', I5)
C      WRITE(lprint ,54557) (LCHROM(JJJJ), JJJJ=1 , MCHRO) 

C    Mutation on the thresholds: nxit1 controls the bit in lchrom
C    where the current subset ARMA structure specification is starting
C    The routine ksarma gives the length of the fragment that identifies
C    the subset ARMA model as klarma, and the bits start from nxit1+1 
C    To prevent out-of-range of the array lchrom, number of threshold
C    parameters is monitored by nthre (nthre < nk)

C        write(lprint, 40519)
C40519 format(1h   , 'mutation on thresholds starts')
C      WRITE(lprint, 54557) (LCHROM(JJJJ), JJJJ=1 , MCHRO) 

      nthre  =  0
      nxit1  =  nxit 
      nbit11  =  NBIT  + 1

      do  4002 i = nbit11, nxit

C  If lchrom(i) = 1 then computes the length of the ARMA model for that
C  regime and add 1 to the threshold parameter counter nthre   
      
      if(lchrom(i) .ge. 1) then

      call ksarma(naux1, nbit,NBIT1,NBIT2, lchrom, mchro, nxit1, klarma)

C      write(lprint, 40660) i , klarma, nxit1
C40660 format(1h   ,'bit n. 'I5, 5X, 'model length ', I5
C     +/1h   ,'current model end-bit nxit1 ', I5)

      nthre  =  nthre  + 1
       if(nthre .ge.  nk)  then

C     too much thresholds: the original chromosome fragment is restored

C       write(lprint, 40217) nthre
C40217  format(1h  ,'No. threshold nthre = ', i5)

       do  47317 ii  =  nbit11  ,  mchro
47317  lchrom(ii)  =   no(kk , ii)
       goto  4001
       endif
      endif 

C        if the bit has to mutate, then the bit lchrom(i) flips

      if(random(l) .le.  p1) then
C        forcing regime deletion for checking 
C      if(lchrom(i) .eq. 1) then
C       WRITE(*,*) 'THE BIT MUTATES'

C      write(lprint, 40661) i , lchrom(i)
C40661 format(1h  ,'bit mutation ', I5, 5X, 'value ', I5)

      ichang = 1
      lchrom(i) =  1  - lchrom(i) 

C  lchrom(i) = 0 implies that the original bit was 1 : the corresponding
C  ARMA model has to be canceled. The number of threshold parameters is
C  decreased by 1. The ARMA model parameters are overwritten with the
C  rest of the cromosome

      if(lchrom(i) .eq. 0) then
      nthre =  nthre - 1
      nxit0 = nxit1 + klarma + 1
      ij =  nxit1 
      do   4003  ii =  nxit0 , mchro
      ij = ij + 1
4003  lchrom(ij) =  lchrom(ii)
      ij1 = ij  +1
      if(ij1 .le.  mchro) then
      do 4091  ij = ij1,  mchro
4091  lchrom(ij)  =   0

      endif

C      write(lprint, 40662) nxit1
C40662 format(1h  ,'chromosome after mutation, nxit1 = ', I5)
C      WRITE(lprint ,54557) (LCHROM(JJJJ), JJJJ=1 , MCHRO) 

C  lchrom(i) = 1 implies that the original bit was 0. The number of
C  threshold parameters is increased by 1. An ARMA model must be
C  generated for the new regime 

      else

      nthre =  nthre +  1

       if(nthre .ge.  nk)  then

C     too much thresholds: the original chromosome fragment is restored

C       write(lprint, 40117) nthre
C40117  format(1h  ,'No. threshold nthre = ', i5)

       do  47309 ii  =  nbit11  ,  mchro
47309  lchrom(ii)  =   no(kk , ii)
       goto  4001
       endif

C        recover in lchro1 the bits that encode the existing model(s)
C        and set lchrom to zero from nxit0 through mchro

      nxit0 = nxit1 + 1
      ij =  0
      do  4006  ii  =  nxit0, mchro
      ij =  ij + 1
4006  lchro1(ij)  =  lchrom(ii)
      ij1 =  ij + 1
      if(ij1 .le.  mchro) then
      do  4116 ii = ij1 , mchro
4116  lchro1(ii)  =  0
      do  40680  ii =  nxit0 , mchro
40680 lchrom(ii)  =   0

      endif

C      write(lprint, 40663) 
C40663 format(1h  ,'bits recovered in lchro1')
C      WRITE(lprint ,54557) (LCHRO1(JJJJ), JJJJ=1 , MCHRO) 

C        generates the AR part of the model for the new regime: first,
C        the AR lags are generated 

       ncp  =  0
       nlagh  =  0
       do  4007 ij = 1 , np
       do 4008  jj = 1 , nbit2
4008   naux1(jj) =  idnint(random(l)) 
       call ndec1(naux1, nbit2, llagh) 
       if(llagh  .le. 0)  goto  4009
       ncp =  ncp  + 1
       nlagh =  nlagh + llagh
       if(ncp .gt. np. or. nlagh .gt. npLAG)  then
       ncp =  ncp - 1
       goto  4009
       else
       naux2(ncp) = nlagh
       endif
4007   continue
4009   continue
       call  intobi(naux1, nbit1, ncp)
       do  4010  jj = 1 , nbit1
       nxit1 = nxit1 + 1
4010   lchrom(nxit1)  =  naux1(jj)

C      WRITE(lprint  ,*) 'NCP ' , NCP
C      WRITE(lprint  ,54557) (LCHROM(JJJJ), JJJJ=1, MCHRO)

C  According to the generated AR lags, the number of parameters is
C  stored	in lchrom. The AR lags are stored next if ncp > 0

       if(ncp. gt .0)  then

C       write(lprint,  *) ncp, ' lags'
C       write(lprint,  40679) (naux2(jjjj), jjjj=1,  ncp)
C40679  format(1h   ,14I5)

       do   4011  ij = 1 ,  ncp

       llagh =  naux2(ij)

C       write(lprint, *) ij, ' lag ', llagh

       call  intobi(naux1, nbit2, llagh)
       do  4012  jj = 1 , nbit2
       nxit1 = nxit1 + 1
4012   lchrom(nxit1)  =  naux1(jj)

C       write(lprint, 40669) nxit1
C40669  format(1h   ,'nxit1 = ', I5, 'binary lag')
C       write(lprint, 40670) (naux1(jjii),  jjii = 1 ,  nbit)
C40670  format(1h   ,70I1)

4011   continue             

       endif

C      WRITE(lprint  ,*) 'MUTATED CHROM., AR PART'
C      WRITE(lprint  ,54557) (LCHROM(JJJJ), JJJJ=1, MCHRO)

C        generates the MA part of the model for the new regime: first,
C        the MA lags are generated 

       ncq  =  0
       nlagh  =  0
       do  4013 ij = 1 , nq
       do 4014  jj = 1 , nbit2
4014   naux1(jj) =  idnint(random(l)) 
       call ndec1(naux1, nbit2, llagh) 
       if(llagh  .le. 0)  goto  4015
       ncq =  ncq  + 1
       nlagh =  nlagh + llagh
       if(ncq .gt. nq. or. nlagh .gt. nqLAG)  then
       ncq =  ncq - 1
       goto  4015
       else
       naux2(ncq) = nlagh
       endif
4013   continue
4015   continue
       call  intobi(naux1, nbit1, ncq)
       do  4016  jj = 1 , nbit1
       nxit1 = nxit1 + 1
4016   lchrom(nxit1)  =  naux1(jj)

C      WRITE(lprint  ,*) 'NCQ ' , NCQ
C      WRITE(lprint  ,54557) (LCHROM(JJJJ), JJJJ=1, MCHRO)

C  According to the generated MA lags, the number of parameters is
C  stored	in lchrom. The MA lags are stored next if ncq > 0

       if(ncq. gt .0)  then

C       write(lprint,  *) ncq, ' lags'
C       write(lprint,  43979) (naux2(jjjj), jjjj=1,  ncq)
C43979  format(1h   ,14I5)

       do   4017  ij = 1 ,  ncq

       llagh =  naux2(ij)

C       write(lprint, *) ij, ' lag ', llagh

       call  intobi(naux1, nbit2, llagh)
       do  4018  jj = 1 , nbit2
       nxit1 = nxit1 + 1
4018   lchrom(nxit1)  =  naux1(jj)

C       write(lprint, 43969) nxit1
C43969  format(1h   ,'nxit1 = ', I5, 'binary lag')
C       write(lprint, 43970) (naux1(jjii),  jjii = 1 ,  nbit)
C43970  format(1h   ,70I1)

4017   continue             

       endif

C      WRITE(lprint  ,*) 'MUTATED CHROM., MA PART'
C      WRITE(lprint  ,54557) (LCHROM(JJJJ), JJJJ=1, MCHRO)

C        add to the chromosome lchrom the rest saved in lchro1

      nxit0 = nxit1 + 1
      ij =  0
      do  4019  ii  =  nxit0, mchro
      ij =  ij + 1
4019  lchrom(ii)  =  lchro1(ij)

C      write(lprint, 40664) nxit1
C40664 format(1h  ,'new model, nxit1 = ', I5)
C      WRITE(lprint ,54557) (LCHROM(JJJJ), JJJJ=1 , MCHRO) 

      endif

C      WRITE(lprint ,*) 'new CHROM., after mutation'
C      WRITE(lprint ,54557) (LCHROM(JJJJ), JJJJ=1, MCHRO)

C        the bit does not mutate, so the counter nxit1 is updated

      else

      IF(LCHROM(I) .GE. 1) THEN
      NXIT1 =  NXIT1 + KLARMA
      ENDIF

      endif

4002  continue  

4001  continue

C  Mutation continues even if none of the first nbit+(n-2mspan) bits
C  changed its value. The routine DECODE checks if the  delay parameter
C  is non zero, number of regimes does not exceed nk and if all regimes
C  contain at least mspan observations. If the previous mutation
C  produced either d=0 (ixflah = -7) or k > nk (ixflah = -1) or some
C  regimes have less than mspan observations (ixflah = -4) then the 
C  chromosome lchrom is restarted 

C      WRITE(lprint, *) 'ENTER DECODE'
C      WRITE(lprint ,54557) (LCHROM(JJJJ), JJJJ = 1 , MCHRO)

      CALL  DECODE(LCHROM, MCHRO, NCHRO, Z, NM, NOB, KKHN, ND, R, MK, NK
     +, NP, NQ, NXP, LAHP, NXQ, LAHQ, NBIT, NBIT1, NBIT2, MSPAN, K, K1
     +, nplag, nqlag, ixflah)

C      write(lprint, 10024) ixflah
C10024 format(1h  ,'decode: ixflah = ', I5)

C       If mutation produced illegal offsprings then the original
C       chromosome is restored
 
        if(ixflah.eq.-7.or.ixflah.eq.-1.or.ixflah.eq.-4)  then
        do   4104 i = 1 , mchro
4104    lchrom(i)  =  no(kk , i)
        endif

C      write(lprint, 54557) (lchrom(jjjj) , jjjj =1 , mchro)

C   Mutation carries on to examine the bits from nxit + 1 to mchro that
C   specify the subset ARMA model structure for each of k regimes
C   The number of regimes k has been decoded by the routine DECODE, the
C   number of threshold parameters is k1 = k - 1
   
      nxit1 =   nxit

      do   4020  ii =  1,  k

C        WRITE(lprint,  40676) ii
C40676 format(1h  , 'REGIME ' , I5, '    LCHROM' )
C        WRITE(lprint, 54557) (LCHROM(JJJJ), JJJJ=1,MCHRO)

C        Mutation of number of AR parameters
C      save the MA part of this regime model and the models belonging
C      to the next regime(s) in lchro1, unless nxit7 greater than mchro

      call hsarma(naux1, nbit, NBIT1,NBIT2,lchrom, mchro, nxit1, llarma)

C        WRITE(lprint, 40674)   llarma
C40674  format(1h  , 'llarma ',  I5)

      nxit7 =  nxit1 + llarma + 1

C        WRITE(lprint, 40675)   nxit7
C40675  format(1h  , 'nxit7 ',  I5)

      if(nxit7 .le.  mchro) then
      ij  =  0
      do  4025  jj  =  nxit7 , mchro
      ij = ij + 1
4025  lchro1(ij) = lchrom(jj)
      ij1 = ij + 1
      do 4026  jj =  ij1, mchro
4026  lchro1(jj)  =   0
      do  40125  jj  =  nxit7 , mchro
40125  lchrom(jj) = 0

C        WRITE(lprint, *) 'AR, SAVE LCHRO1'
C        WRITE(lprint,54557) (LCHRO1(JJJJ), JJJJ=1,MCHRO)
C        WRITE(lprint, *) 'AR, current lchrom'
C        WRITE(lprint,54557) (LCHROM(JJJJ), JJJJ=1,MCHRO)

      endif

C        checking the current number of AR parameters --> ncp 

C        WRITE(lprint, 40671)   Nxit1
C40671  format(1h  , 'Nxit1',  I5)

      nxit0 =  nxit1 +  1
      nxit9 =  nxit1 + nbit1
      ij  =  0
      do  4024  jj = nxit0 , nxit9
      ij =  ij + 1
4024  naux1(ij) =  lchrom(jj)
      call  ndec1(naux1, nbit1, ncp)

C        WRITE(lprint, 40672)   NCP
C40672  format(1h  , 'NCP ',  I5)

C        mutation on the number of AR parameters (bitwise)

      do  4022  jj =  nxit0 , nxit9
      if(random(l) .le. p1) then
      ichang =  1
      lchrom(jj)  =  1 - lchrom(jj)
      endif
4022  continue

C        checking the new number of AR parameters --> ncp1

      ij = 0
      do  4021  jj =  nxit0 , nxit9
      ij =  ij + 1
4021  naux1(ij) =  lchrom(jj)
      call  ndec1(naux1, nbit1, ncp1)

C        WRITE(lprint, 40678)   NCP1
C40678  format(1h  , 'NCP1 ',  I5)

C  Comparison between ncp and ncp1 : if equal, go to mutation of AR lags
C  if not, delete some lags at random if ncp > ncp1, add some lags at
C  random if ncp1 > ncp and then go to mutation of AR lags
 
      if(ncp .eq. ncp1)  then
      nxit5=nxit1+nbit1+ncp*nbit2
      goto  4023
      endif

C        ncp < ncp1 : add some AR lags

      if(ncp .lt. ncp1) then
       DO  33881 iki = 1, np
33881  FZ(iki)  =  zero
       if(ncp.gt.0) then
       do  4027  jjj  =  1 , ncp
       ij = 0
       nxit0 = nxit9 + 1
       nxit9 =  nxit9 + nbit2
       do  4028  jj =  nxit0 , nxit9
       ij =  ij + 1
4028   naux1(ij) =  lchrom(jj)
       call  ndec1(naux1, nbit2, llagh)
       FZ(llagh)  =   1D0
4027   continue 
       endif

C      write(lprint, *) 'Existing lags, FZ'
C      write(lprint, 18024) (FZ(jjii), jjii = 1 , np)
C18024 format(1h   ,14F5.1)

C  The np-length real array FZ now contains 1 in the positions which
C  indicate an active AR lag. New lags are generated at random in (1,np)
C  If some is active already, then go to the next lag. If the lag
C  counter exceeds np, it is reset to 1.
 
       nlag1 = ncp1 - ncp
       do  4088  ij = 1 , nlag1
       llagh = max0(1,min0(np,idnint(1D0+random(l)*dble(np-1))))
       jjj =  llagh-1
4029   jjj  =  jjj + 1
       if(jjj .gt. np) jjj =  1
       if(FZ(jjj) .gt. tol)  goto 4029
       FZ(jjj)  =  1D0
4088   continue

C      write(lprint, *)  'Current lags, FZ'
C      write(lprint, 18025) (FZ(jjii) ,  jjii =1 ,  np)
C18025 format(1h   ,14F5.1)

C  The new AR structure is stored in the chromosome. First, the ones in
C  FZ are transformed in lags in naux2 (in ascendig order automatically)
C  Then, ncp is reset to the new number of AR lags (ncp1) 

       nxit5  =  nxit1
       ncp  =  0
       do  4030 iki = 1 , np
       if(FZ(iki) .gt. tol)  then
       ncp =  ncp + 1
       naux2(ncp) = iki              
       endif
4030   continue

C      write(lprint, *) ncp, 'current lags naux2'
C      write(lprint, 4031) (naux2(jjii), jjii = 1, ncp)
C4031  format(1h   ,14I5)

C  number of AR parameters ncp and parameter lags naux2(1)...naux2(ncp)
C  (these latter if ncp > 0) are stored in lchrom, the current location
C  nxit1 (which excludes as well the number of AR lags) is not altered
C  in view of the next task (mutation of the AR lag values)

       call intobi(naux1, nbit1, ncp)
       nxit0 = nxit5 + 1
       nxit5 = nxit5 + nbit1
       ij = 0
       do 4083  jj =nxit0, nxit5
       ij = ij +  1
4083   lchrom(jj) = naux1(ij)

       if(ncp  .le.  0)  goto  4034
       do 4032 iki = 1 , ncp
       llagh =  naux2(iki)
       call intobi(naux1, nbit2, llagh)
       nxit0 = nxit5 + 1
       nxit5 = nxit5 + nbit2
       ij = 0
       do 4033  jj =nxit0, nxit5
       ij = ij +  1
4033   lchrom(jj) = naux1(ij)
4032   continue
4034   continue

C      write(lprint, *) 'new structure, nxit1 = ', nxit1
C        WRITE(lprint,54557) (LCHROM(JJJJ), JJJJ=1,MCHRO)

C        ncp > ncp1 : delete some AR lags
C        the new number of AR lags is stored in lchrom

      else

       nxit5 =  nxit1
       call intobi(naux1, nbit1, ncp1)
       nxit0 = nxit5 + 1
       nxit5 = nxit5 + nbit1
       ij = 0
       do 4036  jj =nxit0, nxit5
       ij = ij +  1
4036   lchrom(jj) = naux1(ij)

       if(ncp1  .le.  0)  goto  4037

C  If ncp1 is zero, all existing lags are canceled, otherwise the ncp
C  lags are decoded and ncp1 among them will be chosen to survive
 
       do  4038 jjj =1 ,  ncp
       ij = 0
       nxit0 = nxit9 + 1
       nxit9 =  nxit9 + nbit2
       do  4039  jj =  nxit0 , nxit9
       ij = ij + 1
4039   naux1(ij) =  lchrom(jj)
       call  ndec1(naux1, nbit2, llagh)
       naux2(jjj)  = llagh
4038   continue

C  nlag1 lag values are deleted, the remaining will survive ; the loop
C  4040 ensures that the same lag value will not be canceled more than
C  once, if already canceled then is canceled the next one

       nlag1 = ncp - ncp1
       do  4041  iki = 1 , nlag1
       jhcanc = max0(1,min0(ncp,idnint(1D0+random(l)*dble(ncp-1))))
       jjj  = jhcanc - 1
4040   jjj =  jjj + 1      
       if(jjj .gt.  ncp) jjj  =  1
       if(naux2(jjj).gt.0) then
       naux2(jjj) =  0
       else
       goto  4040
       endif
4041   continue

C      write(lprint, *) 'naux2, zero = canceled lag'
C      write(lprint, 74041) (naux2(jjii), jjii = 1 , ncp)
C74041 format(1h   ,14I5)

C        pack the lags which survived and store them in lchrom

       do  4042  iki = 1 , ncp
       if(naux2(iki) .gt. 0) then
       llagh = naux2(iki)
       call intobi(naux1, nbit2, llagh)
       nxit0 = nxit5 + 1
       nxit5 = nxit5 + nbit2
       ij = 0
       do 4043  jj =nxit0, nxit5
       ij = ij +  1
4043   lchrom(jj) = naux1(ij)
       endif
4042   continue
4037   continue

C        ncp is set to the new number of AR lags ncp1

      ncp  =  ncp1

      endif

4023  continue

C        set the lchrom model fragments to zero

      nxit0 =  nxit5 + 1
      if(nxit0  .le.  mchro) then
      do  40681  jjjj  =  nxit0  ,  mchro
40681 lchrom(jjjj)  =   0
      endif 

C      write(lprint, *) 'lchrom ready for subset AR lag mutation'
C        WRITE(lprint,54557) (LCHROM(JJJJ), JJJJ=1,MCHRO)

C       Subset AR lag mutation
C       The current nxit1 location excludes the number of the AR lags,
C       the lag values come next (unless ncp is equal to zero)

      nxit9  =  nxit1 + nbit1

C      write(lprint,*)  'nxit9 ', nxit9
C      write(lprint, *)  'ncp ', ncp

      kjj =  0

      if(ncp  .gt.  0) then

      do  4046 jj = 1 , ncp      
      jckang  = 0
      nxit0 = nxit9 + 1
      nxit9 = nxit9 + nbit2
       ij  =  0
      do 14046 jjjj =  nxit0 , nxit9
       ij =  ij +  1
14046 naux1(ij) =  lchrom(jjjj)

C      write(lprint, *) 'Existing lag n. ', jj
C      write(lprint, 54557) (naux1(jjii), jjii=1,  nbit)

      do  40666 jjjj  = 1 ,  nbit2
      if(random(l) .le. p1) then
       jckang  = 1
      naux1(jjjj) = 1 - naux1(jjjj)    
      endif
40666 continue

      call  ndec1(naux1, nbit2,  llagh)

C      write(lprint, *) 'mutated lag, jckang = ', jckang
C      write(lprint,54557) (naux1(jjii), jjii=1,  nbit)
C      write(lprint,*)  'llagh ', llagh

      if(llagh  .gt.  0  .and.  llagh  .le.  npLAG) then
      kjj  =  kjj  + 1
      fz(kjj)  =  llagh

C      write(lprint , *) 'valid lag ', kjj, fz(kjj)

      endif
4046  continue

      if(jckang  .eq.  1)  ichang =  1

      if(kjj  .gt.  0)  then

      call gradu(fz, fitx, fity, kkn, kng, kjj)

C      write(lprint, *) 'lags: raw, ordered'
C      write(lprint, 94046) (fz(jjii) , jjii=1 , kjj)
C      write(lprint, 94046) (fity(jjii) , jjii=1 , kjj)
C94046 format(1h   ,14F5.1)

      kji =  0
      do 40691  jj = 1 , kjj
       if(jj  .eq.  1)  then
       kji = kji  +  1
       naux2(kji)  =  fity(jj)
       else
       if(dabs(fity(jj) - fity(jj - 1)) .gt. tol)  then
       kji =  kji + 1
       naux2(kji) =  fity(jj)
       endif
       endif
40691 continue
 
      kjj  =  kji
      call  intobi(naux1, nbit1,  kjj)
      nxit0  =  nxit1 +  1
      nxit1  =  nxit1 +  nbit1
       ij =  0
      do  40667  jj =  nxit0 , nxit1
       ij =  ij + 1
40667 lchrom(jj)  =  naux1(ij)
      if(kjj  .gt.  0)  then
      do  40668  jjjj  =  1 ,  kjj
      llagh  =  naux2(jjjj)
      call intobi(naux1, nbit2, llagh)  
      nxit0  =  nxit1 +  1
      nxit1  =  nxit1 +  nbit2
       ij =  0
      do  44669  jj =  nxit0 , nxit1
       ij =  ij + 1
44669 lchrom(jj)  =  naux1(ij)
40668  continue
       endif

C        close : if(kjj .gt. 0)

      else

      nxit0 =  nxit1 + 1
      nxit1 =  nxit1  + nbit1
      do 88899 jjii  =  nxit0, nxit1
88899 lchrom(jjii)  =   0      

      endif

C        close : if(ncp .gt. 0)

      else

      nxit0 =  nxit1 + 1
      nxit1 =  nxit1  + nbit1
      do 99899 jjii  =  nxit0, nxit1
99899 lchrom(jjii)  =   0      

      endif

C        set to zero the rest of the current chromosome lchrom

       nxit7 =  nxit1  + 1 
       if(nxit7 .le.  mchro)  then
      do  40325  jj  =  nxit7 , mchro
40325  lchrom(jj) = 0
       endif

C        WRITE(lprint, 44670)
C44670 format(1h   , 'SUBSET AR MUTATION' )
C        WRITE(lprint ,54557) (LCHROM(JJJJ),JJJJ=1,MCHRO)

C Recover the rest of the chromosome saved in the workspace array lchro1

       nxit0 = nxit1 + 1
       ij = 0
       do  4035  jj = nxit0 , mchro
       ij = ij + 1
4035   lchrom(jj) = lchro1(ij)

C         WRITE(lprint  ,*) 'RECOVER LCHROM'
C        WRITE(lprint  ,54557) (LCHROM(JJJJ), JJJJ=1, MCHRO)

C  Mutation of number of MA parameters
C  Save the models belonging to the remaining regime(s) unless nxit7 is
C  greater than mchro

      call hsarma(naux1, nbit, NBIT1, NBIT2, lchrom,mchro,nxit1, llarma)

C      write(lprint, *) 'llarma ', llarma

      nxit7 =  nxit1 + llarma + 1

C      write(lprint, *) 'nxit7 ', nxit7

      if(nxit7  .le.  mchro)  then
      ij  =  0
      do  4051  jj  =  nxit7 , mchro
      ij = ij + 1
4051  lchro1(ij) = lchrom(jj)
      ij1 = ij + 1
      do 4052  jj =  ij1, mchro
4052  lchro1(jj)  =   0
      do  40225  jj  =  nxit7 , mchro
40225  lchrom(jj) = 0

C        WRITE(lprint  ,*) 'MA, SAVE LCHRO1'
C        WRITE(lprint  ,54557) (LCHRO1(JJJJ),JJJJ=1,MCHRO)
C        WRITE(lprint  ,*) 'MA, current lchrom '
C        WRITE(lprint  ,54557) (LCHROM(JJJJ),JJJJ=1,MCHRO)

      endif

C        checking the current number of MA parameters --> ncq

C        WRITE(lprint,*) 'nxit1 ',  NXIT1

      nxit0 =  nxit1 +  1
      nxit9 =  nxit1 + nbit1
      ij  =  0
      do  4047  jj = nxit0 , nxit9
      ij =  ij + 1
4047  naux1(ij) =  lchrom(jj)
      call  ndec1(naux1, nbit1, ncq)

C      write(lprint, *)  'NCQ ', ncq

C        bitwise mutation on the number of MA parameters 

      do  4048  jj =  nxit0 , nxit9
      if(random(l) .le. p1) then
      ichang =  1
      lchrom(jj)  =  1 - lchrom(jj)
      endif
4048  continue

C        checking the new number of MA parameters --> ncq1

      ij = 0
      do  4049  jj =  nxit0 , nxit9
      ij =  ij + 1
4049  naux1(ij) =  lchrom(jj)
      call  ndec1(naux1, nbit1, ncq1)

C      write(lprint, *)  'NCQ1 ', ncq1

C  Compare ncq and ncq1 : if equal, go to the mutation of MA lag values,
C  otherwise delete some lags if ncq1 < ncq, add some lags if ncq < ncq1

      if(ncq .eq. ncq1)  then  
      nxit5 = nxit1 +  nbit1 + ncq * nbit2
      goto  4050
      endif

C        ncq < ncq1 : add some MA lags at random

      if(ncq .lt. ncq1) then
       DO  4053 iki = 1, nq
4053   FZ(iki)  =  zero
       if(ncq.gt.0) then
       do  4054  jjj  =  1 , ncq
       ij = 0
       nxit0 = nxit9 + 1
       nxit9 =  nxit9 + nbit2
       do  4055  jj =  nxit0 , nxit9
       ij =  ij + 1
4055   naux1(ij) =  lchrom(jj)
       call  ndec1(naux1, nbit2, llagh)
       FZ(llagh)  =   1D0
4054   continue 
       endif

C      write(lprint,  *) 'Existing lags FZ'
C      write(lprint, 19024) (FZ(jjii)  , jjii = 1,  nq)
C19024 format(1h    ,14F5.1)

C  The nq-length real array FZ now contains 1 in the positions which
C  indicate an active MA lag. New lags are generated at random in (1,nq)
C  If some is active already, then go to the next lag. If the lag
C  counter exceeds nq, it is reset to 1.

       nlag1 = ncq1 - ncq
       do  4056  ij = 1 , nlag1
       llagh = max0(1,min0(nq,idnint(1D0+random(l)*dble(nq-1))))
       jjj =  llagh-1
4057   jjj  =  jjj + 1
       if(jjj .gt. nq) jjj =  1
       if(FZ(jjj) .gt. tol)  goto 4057
       FZ(jjj)  =  1D0
4056   continue

C      write(lprint,  *) 'Current lags FZ'
C      write(lprint, 19025) (FZ(jjii)  , jjii = 1,  nq)
C19025 format(1h    ,14F5.1)

C  The new MA structure is stored in the chromosome. First, the ones in
C  FZ are transformed in lag in naux2 (in ascending order automatically)
C  Then ncq is reset to the new number of MA lags (ncq1)

       nxit5 =  nxit1
       ncq  =  0
       do  4058 iki = 1 , nq
       if(FZ(iki) .gt. tol)  then
       ncq =  ncq + 1
       naux2(ncq) = iki              
       endif
4058   continue

C       write(lprint, *) ncq , ' current lags naux2'
C       write(lprint, 4059) (naux2(jjii), jjii = 1  , ncq)
C4059   format(1h  ,  14I5)

C    number of MA parameters ncq and parameter lags are stored in lchrom
C    the current location nxit1 is not altered in view of the next task,
C    that is the (possible) mutation of the MA lags values
 
       call intobi(naux1, nbit1, ncq)
       nxit0 = nxit5 + 1
       nxit5 = nxit5 + nbit1
       ij = 0
       do 4060  jj =nxit0, nxit5
       ij = ij +  1
4060   lchrom(jj) = naux1(ij)

       if(ncq  .le.  0)  goto  4061
       do 4062 iki = 1 , ncq
       llagh =  naux2(iki)
       call intobi(naux1, nbit2, llagh)
       nxit0 = nxit5 + 1
       nxit5 = nxit5 + nbit2
       ij = 0
       do 4063  jj =nxit0, nxit5
       ij = ij +  1
4063   lchrom(jj) = naux1(ij)
4062   continue
4061   continue

C      write(lprint, *) 'new structure, nxit1 = ', nxit1
C        WRITE(lprint, 54557) (LCHROM(JJJJ),JJJJ=1,MCHRO)

C        ncq > ncq1 : delete some MA lags,
C        the new number of MA lags is stored in lchrom

      else

       nxit5 =  nxit1
       call intobi(naux1, nbit1, ncq1)
       nxit0 = nxit5 + 1
       nxit5 = nxit5 + nbit1
       ij = 0
       do 4066  jj =nxit0, nxit5
       ij = ij +  1
4066   lchrom(jj) = naux1(ij)

       if(ncq1  .le.  0)  goto  4067

C  If ncq1 is zero all existing lags are cenceled, otherwise the ncq
C  lags are decoded and ncq1 among them will be chosen to survive

       do  4068 jjj =1 ,  ncq
       ij = 0
       nxit0 = nxit9 + 1
       nxit9 =  nxit9 + nbit2
       do  4069  jj =  nxit0 , nxit9
       ij = ij + 1
4069   naux1(ij) =  lchrom(jj)
       call  ndec1(naux1, nbit2, llagh)
       naux2(jjj)  = llagh
4068   continue

C  nlag1 lag values are deleted, the remaining will survive ; the loop
C  4070 ensures that the same lag value will not be canceled more than
C  once, if already canceled then is canceled the next one

       nlag1 = ncq - ncq1
       do  4071  iki = 1 , nlag1
       jhcanc = max0(1,min0(ncq,idnint(1D0+random(l)*dble(ncq-1))))
       jjj  = jhcanc - 1
4070   jjj =  jjj + 1      
       if(jjj .gt.  ncq) jjj =  1
       if(naux2(jjj).gt.0) then
       naux2(jjj) =  0
       else
       goto  4070
       endif
4071   continue

C      write(lprint, *) 'naux2, zero = canceled lag'
C      write(lprint, 74941) (naux2(jjii), jjii = 1 , ncq)
C74941 format(1h  ,  14I5)

C        pack the lags which survived and store them in lchrom

       do  4072  iki = 1 , ncq
       if(naux2(iki) .gt. 0) then
       llagh = naux2(iki)
       call intobi(naux1, nbit2, llagh)
       nxit0 = nxit5 + 1
       nxit5 = nxit5 + nbit2
       ij = 0
       do 4073  jj =nxit0, nxit5
       ij = ij +  1
4073   lchrom(jj) = naux1(ij)
       endif
4072   continue
4067   continue

C        ncq is set to the new number of MA lags ncq1

       ncq =  ncq1

      endif

4050  continue

C        set the lchrom model fragments to zero

       nxit0 = nxit5 + 1
      if(nxit0  .le.  mchro)  then
      do  49681  jjjj =  nxit0 , mchro
49681 lchrom(jjjj)  =  0
      endif

C      write(lprint, *) 'lchrom ready for subset MA lag mutation'
C        WRITE(lprint, 54557) (LCHROM(JJJJ),JJJJ=1,MCHRO)

C       Subset MA lag mutation
C       The current nxit1 location excludes the number of the AR lags,
C       the AR lag values, the number of MA lags; the MA lag values come
C       next (unless ncq is equal to zero)

      nxit9  =  nxit1  +  nbit1

C      write(lprint, *) 'nxit9 ' , nxit9
C      write(lprint, *) 'ncq ', ncq

      kjj  =  0

      if(ncq  .gt.  0)  then

      do  4076 jj = 1 , ncq      
      jckang  = 0
      nxit0 = nxit9 + 1
      nxit9 = nxit9 + nbit2
      ij  =  0
      do 14076  jjjj =  nxit0 , nxit9
      ij  =  ij +  1
14076 naux1(ij) =  lchrom(jjjj)

C      write(lprint, *) 'Existing lag n. ' , jj
C      write(lprint, 54557) (naux1(jjii), jjii = 1 , nbit)

      do  49666  jjjj  =  1  ,  nbit
      if(random(l) .le. p1) then
      jckang = 1
      naux1(jjjj) = 1 - naux1(jjjj)
      endif
49666 continue

      call  ndec1(naux1, nbit2, llagh)

C      write(lprint, *) 'mutated lag, jckang = ' , jckang
C      write(lprint, 54557) (naux1(jjii), jjii = 1 , nbit)
C      write(lprint, *) 'llagh ',  llagh

      if(llagh  .gt. 0  .and.  llagh .le.  nq) then
      kjj  =  kjj + 1
      fz(kjj) =  llagh

C      write(lprint, *) 'valid lag ', kjj, fz(kjj)

      endif
4076  continue

      if(jckang  .eq.  1)  ichang =  1

      if(kjj .gt.  0)  then

      call  gradu(fz, fitx, fity, kkn, kng, kjj)

C      write(lprint, *)  'lags: raw, ordered'
C      write(lprint, 99046) (fz(jjii), jjii =1 , kjj)
C      write(lprint, 99046) (fity(jjii), jjii =1 , kjj)
C99046 format(1h  ,  14F5.1)

      kji  =  0
      do  99691  jj =  1 ,  kjj
       if(jj .eq. 1)  then
       kji =  kji + 1
       naux2(kji) =  fity(jj)
       else
       if(dabs(fity(jj) - fity(jj - 1))  .gt.  tol)  then
       kji =  kji  + 1
       naux2(kji) =  fity(jj)
       endif
       endif
99691 continue

      kjj  =  kji
      call  intobi(naux1, nbit1, kjj)
      nxit0 =  nxit1 + 1
      nxit1 =  nxit1  +  nbit1
       ij  =  0
      do  99667 jj  =  nxit0 , nxit1
       ij  =  ij + 1
99667 lchrom(jj) = naux1(ij)
      if(kjj .gt.  0)  then
      do  99668  jjjj = 1 ,  kjj
      llagh = naux2(jjjj)
      call intobi(naux1, nbit2, llagh)
      nxit0 =  nxit1 + 1
      nxit1 =  nxit1  +  nbit2
       ij =  0
      do  99669 jj =  nxit0 , nxit1
       ij = ij  + 1
99669 lchrom(jj) =  naux1(ij)
99668 continue
      endif

C        close : if(kjj .gt. 0)

      else

      nxit0 =  nxit1 + 1
      nxit1 =  nxit1  + nbit1
      do 88898 jjii  =  nxit0, nxit1
88898 lchrom(jjii)  =   0      

      endif

C        close : if(ncq .gt. 0)

      else

      nxit0 =  nxit1 + 1
      nxit1 =  nxit1  + nbit1
      do 99898 jjii  =  nxit0, nxit1
99898 lchrom(jjii)  =   0      

      endif

C        set to zero the rest of the current chromosome lchrom

      nxit7 =  nxit1  + 1
      if(nxit7  .le.  mchro)  then
      do  99325  jj =  nxit7 , mchro
99325 lchrom(jj)  =  0
      endif

C        WRITE(lprint, *) 'SUBSET MA MUTATION'
C        WRITE(lprint, 54557) (LCHROM(JJJJ),JJJJ=1,MCHRO)

C recover the rest of the chromosome saved in the workspace array lchro1

       nxit0 = nxit1 + 1
       ij = 0
       do  4075  jj = nxit0 , mchro
       ij = ij + 1
4075   lchrom(jj) = lchro1(ij)

C        WRITE(lprint, *) 'RECOVER LCHROM'
C        WRITE(lprint , 54557) (LCHROM(JJJJ),JJJJ=1,MCHRO)

4020  continue 
 
C        Computes the fitness function: the chromosome is accepted 
C        unless ixflah is different from zero

      FFFIT =  FFB(lchrom, mchro, z, mob, nob, nm, kkhn, kkkn, nd
     +, r, MK, nk, k, nj, s2, nbit, mx, nx, NP, NQ, nxp, lahp, nxq, lahq  
     +,icri,crki,xeta, X, RES, VFIT, XMAT, PHI, AUX1, VC1, BB, CC, IP,IQ
     +, MSPAN, nchro, NBIT1,NBIT2,nplag,nqlag
     +, NHP0, NHP1, NHP,iprint, lprint, ixflah)

C       write(lprint,  30301) ixflah
C30301  format(1h  ,  'ffb : ixflah = ', I5)

       if(ixflah.eq.0) then
      fitn(kk)  = fffit
      do  4005  i =  1  ,  mchro
4005  no(kk , i)  =  lchrom(i)
       endif

C       write(lprint, 54557) (lchrom(i),  i=1, mchro)

4000    CONTINUE

C      write(lprint, 129) 
C      WRITE(LPRINT,  55596)  
C55596  FORMAT(1H  ,'Mutation ')
C       do 76044  kk  =  1  ,   nl
C76044  write(lprint, 54557) (no(kk , j),  j=1, mchro)
C      write(lprint, 129) 
C129    format(1h    ,'//'  )

	GOTO  1234

C         end   iterations

2100    CONTINUE

C       iprint  =  1
         IPRINT  = 0

      FBEST =  FFB(lbest, mchro, z, mob, nob, nm, kkhn, kkkn, nd
     +, r, MK, nk, k, nj, s2, nbit, mx, nx, NP, NQ, nxp, lahp, nxq, lahq  
     +,icri,crki,xeta, X, RES, VFIT, XMAT, PHI, AUX1, VC1, BB, CC, IP,IQ
     +, MSPAN, nchro, NBIT1,NBIT2,nplag,nqlag
     +, NHP0, NHP1, NHP,iprint, lprint, ixfla1)
       
	     iter1  =  iter  - 1
C	     write(LPRINT ,  55570)  iter1, ITER2
C55570   format(1x/4x,'End genetic algorithm : iter, iter2 = ', 2i10)
C       write(lprint,  30311) ixfla1
C30311  format(1h  ,  'ixfla1 = ', I5)
C	  WRITE(LPRINT,5557)   NBEST  , FBEST
C5557     FORMAT(1X,4X,'Migliore cromosoma ',I4,2X,'Fitn.Funct.',E12.5)
C       WRITE(LPRINT,5556)(lbest(i), I=1,mchro)
C5556    format(1h   ,70I1)
         ssq  = zero
          do  55  i =1 ,  k
55        ssq  =  ssq +  s2(i)
C	     write(LPRINT ,  10009)  ssq, crki
C10009     format(1h  ,'Pooled res. sum of squares, criterion ', 2D15.6)

       IF(CRKI.  LT.  CRKIQ)  THEN
       CRKIQ =  CRKI
        DO  47471  I =1  ,  MCHRO
47471  LKYROM(I)  =  LBEST(I)
        ENDIF

47470    CONTINUE

       RETURN

        END


      subroutine  intobi(m, maxout, n)

C    Converte il numero intero n nella stringa binaria m lunga maxout

      DIMENSION  m(maxout)
      do  3  i  =1  ,   maxout
3     m(i)  =  0
      kc  =  n
      i  =  maxout + 1
1     continue
      jc  =  kc
      kc  =  jc / 2
      iresto =  jc - 2 * kc
      i = i - 1
      m(i) = iresto
      if(kc .gt. 0)  goto  1
2     continue
      return
      end


        subroutine NDEC1(m,maxout,n)

C       Converte in numero intero n la stringa binaria m lunga maxout
        
        DIMENSION m(maxout)
        n =  0
        DO 3  I  = 1 ,  MAXOUT
3       n = n + m(i) * (2 ** (MAXOUT - I))
         RETURN
         END


      subroutine  ksarma(naux1, nbit, NBIT1,NBIT2,lchrom, mchro, nxit1
     +, klarma)

c  Workspace routine for GTARMA: for any given regime, finds the subset
c  ARMA binary code structure recorded starting from the bit nxit1+1 in
c  the binary array lchrom and stores its length in klarma 

      dimension  naux1(nbit),  lchrom(mchro)

c  initializes the subset ARMA structure length l1

      l1 =  0

c finds the length of the chromosome fragment that specifies the AR part

      ih1  =  nxit1 + 1
      ih2  =   nxit1 +  nbit1
      ij  =  0
      do  35  iki  =  ih1 , ih2
      ij  =  ij  + 1
35    naux1(ij)  =  lchrom(iki)
      call  ndec1(naux1, nbit1, nyp)
      l1  =  l1 + nbit1 + nbit2 * nyp

c finds the length of the chromosome fragment that specifies the MA part

      ih1 =  nxit1 + l1 + 1
      ih2  =  nxit1 + l1 + nbit1  
      ij  =  0
      do  36  iki  =  ih1 , ih2
      ij  =  ij  + 1
36    naux1(ij)  =  lchrom(iki)
      call  ndec1(naux1, nbit1, nyq)
      l1  =  l1 + nbit1 + nbit2 * nyq

c   the resulting length l1 is stored as output in klarma

      klarma =  l1

      return
      end


      subroutine  hsarma(naux1, nbit,NBIT1,NBIT2, lchrom, mchro, nxit1
     +, klarma)

c  Workspace routine for GTARMA: for any given regime, finds the subset
c  either AR or MA binary code structure recorded starting from the bit
c  nxit1+1 in the binary array lchrom and stores its length in klarma 

      dimension  naux1(nbit),  lchrom(mchro)

c        initializes the subset either AR or MA structure length l1

      l1 =  0

c        finds the length of the chromosome fragment that specifies
c        either the AR or MA part of the model

      ih1  =  nxit1 + 1
      ih2  =   nxit1 +  nbit1
      ij  =  0
      do  35  iki  =  ih1 , ih2
      ij  =  ij  + 1
35    naux1(ij)  =  lchrom(iki)
      call  ndec1(naux1, nbit1, nyp)
      l1  =  l1 + nbit1 + nbit2 * nyp

c        the resulting length l1 is stored as output in klarma

      klarma =  l1

      return
      end


      subroutine  gcheck(lchrom, nchro, npd, nk, mspan, ifault)

C       gcheck controlla che per ciascun regime vi siano almeno mspan 
C       osservazioni; gli errori possibili sono:
C       ifault = -1: tra la coda iniziale e finale non ci sono istanti
C       ifault = -2: almeno un regime ha meno di mspan osservazioni
C       ifault = -3: il cromosoma codifica piu' regimi del massimo nk
 
      dimension   lchrom(nchro)
      ifault  =  0
      npd1  =  npd +  1
      npd2  =  nchro - npd
      if(npd2  .lt.  npd1)  then
      ifault  =  -1
      return
      endif
      i1  = 0
      ic  =  0
      do  1   i  =  npd1  , npd2
      if(lchrom(i) .le.  0)  then
      ic =  ic +  1
c      write(*,*) 'i', i, lchrom(i), ic
      else
c      write(*,*) 'i', i , lchrom(i), ic
       if(ic  .le.  mspan) then
       ifault  =  -2
       return
       else
       i1 = i1  + 1
       ik  =  i
       ic  =  1
c      write(*,*) 'i1,ik,ic', i1 , ik, ic
       endif
      endif
1     continue
      if(npd2 - ik + 1  .lt.  mspan)  then
      ifault  =  -2
      return
      endif
      if(i1  .gt.  nk - 1)  then
      ifault  =  -3
      return
      endif
      return
      end
      

      SUBROUTINE PWEIG(PHI, THETA, XPI, NP, NQ, N)
        IMPLICIT REAL*8(A-H , O-Z)

C      PI-WEIGHTS COMPUTATION : BOX-JENKINS (1976) P. 102

      DIMENSION  PHI(NP), THETA(NQ) , XPI(N)

      XPI(1)  =  PHI(1) +  THETA(1)
      IF(N  .LE.  1)  RETURN
      DO  10  I  = 2 ,  N
      XPI(I) =  PHI(I)  +  THETA(I)
      I1  =  I - 1
      DO  11 J = 1  , I1
      JJ  = I1 -  J + 1
      XPI(I) = XPI(I) - XPI(J) * THETA(JJ)
11    CONTINUE
10    CONTINUE

      RETURN

      END


        SUBROUTINE  RESID(Z , NDR, PA, NP, R, VFIT, SSQ, IEXIT, imodel)
        IMPLICIT REAL*8(A-H , O-Z)
        DIMENSION Z(NDR), PA(NP), R(NDR), VFIT(NDR)                             
C      COMMON/THAR/ icri, nkp, k, k1, nd, rk(10), nnp(10), ppa(10,11)                                     
       COMMON/THAR/ icri, nkp, k, k1, nd, nnp(10)
       COMMON/THAR1/ rk(10), ppa(10,11)
       data   tol / 8.43D-37 /                                          
       data     big  / 3.37D+38 /                                       
        IEXIT = 0
                                                                                
C      REAL ARRAY R IS SET TO ZERO                                      
                                                                        
        DO 1 I=1,NDR                                                    
1       R(I) = 0.0                                                      

C      PREDICTED TIME SERIES ACCORDING TO THE AUTOREGRESSIVE MODEL

         if(imodel .eq. 1)  then
        if(np .lt. 1)  goto 1402
        do  1401 i=1, np
1401    r(i)  =  z(i)
1402    continue
         np1 =  np + 1
        DO  1400  I  =  np1 ,  NDR
        DO  2400  J  =  1  ,  NP
2400    R(I) = R(I) + PA(J) * Z(I - J)
1400     CONTINUE
          endif

C      PREDICTED TIME SERIES ACCORDING TO THE STATE DEPENDENT EAR MODEL
 
         if(imodel .eq. 2)  then
         G  =  PA(NP)
         NQ1 = (NP - 1) / 2
        do  140 i=1, nq1
140     r(i)  =  z(i)
         nq2 =  nq1 + 1
        DO 100   I =  nq2 ,NDR
        DO  110  J  =  1 ,  NQ1
        J1  =  NQ1 + J
        C  =  PA(J) + PA(J1) * DEXP(-G * (Z(I - 1) ** 2))
        R(I) = R(I) +  C * Z(I - J)
       IF(DABS(R(I)) .LT. TOL) R(I) = 0.0
      IF(DABS(R(I)).GT.BIG) THEN
      IEXIT = -1
      RETURN
      ENDIF
110    CONTINUE
100     continue
          endif

C      PREDICTED TIME SERIES ACCORDING TO THE BILINEAR MODEL

         if(imodel .eq. 3)  then
        do  3401 i=1, np
3401    r(i)  =  z(i)
         np1 =  np + 1
        DO  3400  I  =  np1 ,  NDR
        R(I)  =  Z(I)
3400     CONTINUE
          endif

C      PREDICTED TIME SERIES ACCORDING TO THE THRESHOLD AR MODEL

        if(imodel .eq. 4)  then

C the first nq1=max(nd,p1,...,pk) predicted values equal the true values

       nq1 = nd
        do  441 i=1  ,  k
       if(nnp(i). gt. nq1) nq1 =  nnp(i)
441     continue
        do  440 i=1, nq1
440     r(i)  =  z(i)

C the nq1+1,...,ndr predicted values are computed from the TAR estimates

         nq2 =  nq1 + 1
      DO  400 I  =  nq2  ,  NDR

C       regime choice depends on Z(t - nd)

        jj =  i - nd
       do 411 iiii =  1 ,  k1
       if(z(jj) .lt. rk(iiii)) goto  414
411    continue
       iiii  =  k
414     continue

       ntp1  =  nnp(iiii)
       ntp  =  nnp(iiii) + 1
       do  412  j =1  ,  ntp
412    pa(j)  =  ppa(iiii ,  j)

C      the regime is k=iiii, the predicted value is inizialized to pa(0)

c      R(I)  =  pa(1)
       R(I)  =  0D0       
      DO  410  J  =  1  ,  Ntp1
      j1  =  j + 1
      R(I) =  R(I) + PA(J1) * Z(I - J)
410   CONTINUE
400   CONTINUE
         endif

C      RESIDUALS COMPUTATION: OBSERVED/SIMULATED TIME SERIES DATA Z 

        DO 101  I = 1 , NDR
        VFIT(I)  =   R(I)
101     R(I)  = Z(I) - R(I)

C      SUM OF SQUARES COMPUTATION

        SSQ  =  DOTPRD(NDR , R , R)
        RETURN                                                          
        END                                                             


       subroutine THCTRL(parver, NQ, icontr, imodel, XCOEFF, XROOTS)
        implicit real*8 (a-h, o-z)
      COMPLEX  XCOEFF(NQ+1), XROOTS(NQ)
      LOGICAL polish
        dimension parver(NQ)
C      COMMON/THAR/ icri, np, k, k1, nd, r(10), nnp(10), ppa(10,11)                                     
       COMMON/THAR/ icri, np, k, k1, nd, nnp(10)
       COMMON/THAR1/ r(10), ppa(10,11)
      data  small/ 8.43D-37/
       lprint  =  6
      ONE =  1D0
c  check AR roots 1-p1B-p2B^2... all outside the unit circle
        if(imodel .eq. 1) then
        icontr  =  -1
      XCOEFF(1) =CMPLX(one)
      DO  1400  J  =  1  ,  NQ
      RCOEFF =  -parver(J)
      JJ  =  J + 1
      XCOEFF(JJ) =CMPLX(RCOEFF)
1400  continue
c      write(lprint, *) (xcoeff(j), j=1 , nq+1)
      polish=.false.
      call zroots(XCOEFF,NQ,XROOTS,polish)
c      write(lprint, *) (xroots(j), j=1 , nq)
      do  1015 i  =  1  ,  NQ
      xnorm = CABS(xroots(i))
c      write(lprint,'(1x,i4,4x, F10.4)') i, xnorm
      if(xnorm .le. one) goto  1409  
1015  continue
       icontr  =  0
1409   continue
        endif
c  check Haggan-Ozaki conditions (Biometrika, 1981)
        if(imodel .eq. 2) then
        icontr =  -1
       RCOEFF =  parver(NQ)
        if(RCOEFF. le. 0D0) goto  2409         
        nq1 = (nq - 1) / 2
        nq2 =  nq1 + 1
      XCOEFF(nq2) =CMPLX(one)
      DO  2400  J  =  1  ,  NQ1
      RCOEFF =  -parver(J)
      JJ  =  NQ1 - J + 1
      XCOEFF(JJ) =CMPLX(RCOEFF)
2400  continue
      write(LPRINT, *) (xcoeff(j), j=1 , nq2)
      polish=.false.
      call zroots(XCOEFF,NQ1,XROOTS,polish)
      write(LPRINT, *) (xroots(j), j=1 , nq1)
      do  2015 i  =  1  ,  NQ1
      xnorm = CABS(xroots(i))
      write(LPRINT,'(1x,i4,4x, F10.4)') i, xnorm
      if(xnorm .ge. one) goto  2409  
2015  continue
      XCOEFF(nq2) =CMPLX(one)
      DO  2401  J  =  1  ,  NQ1
      J1 = J + NQ1
      RCOEFF =  -(parver(J) + parver(J1))
      JJ  =  NQ1 - J + 1
      XCOEFF(JJ) =CMPLX(RCOEFF)
2401  continue
      write(LPRINT, *) (xcoeff(j), j=1 , nq2)
      polish=.false.
      call zroots(XCOEFF,NQ1,XROOTS,polish)
      write(LPRINT, *) (xroots(j), j=1 , nq1)
      inside  = 0
      do  2025 i  =  1  ,  NQ1
      xnorm = CABS(xroots(i))
      write(LPRINT,'(1x,i4,4x, F10.4)') i, xnorm
      if(xnorm .lt. one) inside = inside + 1  
2025  continue
c      pause
      if(inside .eq. nq1) goto 2409
      sum = 0D0
      DO  2402  J  =  1  ,  NQ1
2402  sum  =  sum +  parver(J)
      sam =  0D0
      DO  2403  J  =  1  ,  NQ1
      J1 = NQ1 + J
2403  sam  =  sam +  parver(J1)
      IF(DABS(SAM)  .LE. small)  goto  2409
      FF =  (1D0 - sum) / sam
      write(LPRINT, 100) sum, sam, FF
100   format(1h  ,'Sum phi = ', F10.4,2x, 'Sum pi = ', F10.4
     +, 2x, 'FF =', F10.4)
c      pause
      if(FF .ge. 0D0  .and. FF .le. 1D0) goto  2409
       icontr = 0 
2409   continue
         endif
c  check bilinear model patrameters
        if(imodel .eq. 3) then
       icontr = 0 
         endif
c  checks that the AR roots p0-p1B-p2B^2... of the k - regimes of the
c  TAR threshold models all lie outside the unit circle
        if(imodel .eq. 4) then
        icontr  =  -1
       do  4401  iiii  =  1 , k
        nq1 = nnp(iiii) 
        nq2 = nnp(iiii) + 1
      XCOEFF(1) =CMPLX(one)
      if(NQ2 .ge. 2)  then
      DO  4400  J  =  2  ,  NQ2
      RCOEFF =  -ppa(iiii , J)
      XCOEFF(J) =CMPLX(RCOEFF)
4400  continue
       endif
      write(lprint, *) (xcoeff(j), j=1 , nq2)
      polish=.false.
      call zroots(XCOEFF,NQ1,XROOTS,polish)
      write(lprint, *) (xroots(j), j=1 , nq1)
      do  4015 i  =  1  ,  NQ1
      xnorm = CABS(xroots(i))
      write(lprint,'(1x,i4,4x, F10.4)') i, xnorm
      if(xnorm .le. one) goto  4409  
4015  continue
4401  continue
       icontr  =  0
4409   continue
         endif
        return
         end


      SUBROUTINE zroots(a,m,roots,polish)
C     roots of a polynomial by Laguerre's method with deflation 
      INTEGER m,MAXM
      REAL EPS
      COMPLEX a(m+1),roots(m)
      LOGICAL polish
      PARAMETER (EPS=1.e-6,MAXM=101)
CU    USES laguer
      INTEGER i,j,jj,its
      COMPLEX ad(MAXM),x,b,c
      do 11 j=1,m+1
        ad(j)=a(j)
11    continue
      do 13 j=m,1,-1
        x=cmplx(0.,0.)
        call laguer(ad,j,x,its)
        if(abs(aimag(x)).le.2.*EPS**2*abs(real(x))) x=cmplx(real(x),0.)
        roots(j)=x
        b=ad(j+1)
        do 12 jj=j,1,-1
          c=ad(jj)
          ad(jj)=b

          b=x*b+c
12      continue
13    continue
      if (polish) then
        do 14 j=1,m
          call laguer(a,m,roots(j),its)
14      continue
      endif
      do 16 j=2,m
        x=roots(j)
        do 15 i=j-1,1,-1
          if(real(roots(i)).le.real(x))goto 10
          roots(i+1)=roots(i)
15      continue
        i=0
10      roots(i+1)=x
16    continue
      return
      END


      SUBROUTINE laguer(a,m,x,its)
C     roots of a polynomial by Laguerre's method  
      INTEGER m,its,MAXIT,MR,MT
      REAL EPSS
      COMPLEX a(m+1),x
      PARAMETER (EPSS=2.e-7,MR=8,MT=10,MAXIT=MT*MR)
      INTEGER iter,j
      REAL abx,abp,abm,err,frac(MR)
      COMPLEX dx,x1,b,d,f,g,h,sq,gp,gm,g2
      SAVE frac
      DATA frac /.5,.25,.75,.13,.38,.62,.88,1./
      do 12 iter=1,MAXIT
        its=iter
        b=a(m+1)
        err=abs(b)
        d=cmplx(0.,0.)
        f=cmplx(0.,0.)
        abx=abs(x)
        do 11 j=m,1,-1
          f=x*f+d

          d=x*d+b
          b=x*b+a(j)
          err=abs(b)+abx*err
11      continue
        err=EPSS*err
        if(abs(b).le.err) then
          return
        else
          g=d/b
          g2=g*g
          h=g2-2.*f/b
          sq=sqrt((m-1)*(m*h-g2))
          gp=g+sq
          gm=g-sq
          abp=abs(gp)
          abm=abs(gm)
          if(abp.lt.abm) gp=gm
          if (max(abp,abm).gt.0.) then
            dx=m/gp
          else
            dx=exp(cmplx(log(1.+abx),float(iter)))

          endif
        endif
        x1=x-dx
        if(x.eq.x1)return
        if (mod(iter,MT).ne.0) then
          x=x1
        else
          x=x-dx*frac(iter/MT)
        endif
12    continue
      pause 'too many iterations in laguer'
      return
      END


      SUBROUTINE GRADU(X,XX,XORD,N,NG,NUM)
      IMPLICIT REAL*8(A - H , O -Z)
      DIMENSION X(NUM),XX(NUM),XORD(NUM),N(NUM),NG(NUM)
          big  =   3.37D+38
      DO   1   I   =  1  , NUM
1     XX(I)   =    X(I)
       K   =    0
3     K  =   K   +   1
      AMIN   =   big
	 DO  2   I   =   1   ,   NUM
       IF(XX(I).GE.AMIN)    GOTO  2
      AMIN   =    XX(I)
       J   =   I
2     CONTINUE
      XORD(K)   =    XX(J)
      N(J)   =   K
      NG(K)   =   J
      XX(J)    =     big
      IF(K.LT.NUM)     GOTO  3
      RETURN
      END


      REAL*8 FUNCTION FFB(lchrom, mchro, z, mob, nhb, nm, kkhn, kkkn, nd
     +, r, MK, nk, k, nj, s2, nbit, mx, nx, NP, NQ, nxp, lahp, nxq, lahq  
     +, icri, crki, c, X, RES, VFIT, XMAT, PHI, AUX1, VC1, BB, CC, IP,IQ
     +, MSPAN, nchro, NBIT1, NBIT2, nplag,nqlag
     +, NHP0, NHP1, NHP, iprint, lprint, ifault)
      IMPLICIT REAL*8 (A - H , O - Z)
      DIMENSION LCHROM(MCHRO), z(nhb), kkhn(nhb), kkkn(nhb), r(nk)
     +, nj(nk), s2(nk), X(nhb), RES(nhb), VFIT(nhb), XMAT(MOB, nX)
     +, PHI(nX), AUX1(nX), VC1(MX, nX), BB(nX), CC(nX), IP(nX), IQ(nX)
     +, nhp(nk), nxp(nk), lahp(MK, np), nxq(nk), lahq(MK, nq)
      COMMON /PPPP/ PPP(5,13)
      data  tol/ 8.43D-37/
       data     big  / 3.37D+38 /                                       

C        NP, NQ ARE THE MAX NUMBER OF PARAMATERS AR, MA
C        NPLAG, NQLAG ARE THE MAX NUMBER OF LAGS AR, MA
C        NHP0, NHP1 ARE THE LOWER AND UPPER BOUNDS WHERE IS TO BE FOUND
C        THE OPTIMUM AR ORDER P FOR THE LONG AUTOREGRESSION IN EACH
C        REGIME, THE RESULT IS STORED IN NHP(II), II=1, ..., NK
C        NX IS TO BE SET TO NHP1 + 1 (FROM OUTSIDE) BUT IT IS USEFUL
C        ONLY FOR THE DIMENSIONS AND THE MAX NUMBER OF PARAMETERS 

c      write(lprint, 152) 
c152   format(1h   ,'Enter FFB  ')

c        set initial value for FFB

      ffb  =  0.0D-000

c        set error flag : ifault = 0 if normal exit

      IFAULT  =  0

c +++    decoding chromosome starts

      CALL  DECODE(LCHROM, MCHRO, NCHRO, Z, NM, NHB, KKHN, ND, R, MK, NK
     +, NP, NQ, NXP, LAHP, NXQ, LAHQ, NBIT, NBIT1, NBIT2, MSPAN, K, K1
     +, nplag, nqlag, IFAULT)

c       number of regimes greater than the maximum allowed nk: IFAULT=-1 
c       either maximum AR lag or AR order greater than np: IFAULT=-2
c       either maximum MA lag or MA order greater than nq: IFAULT=-3
c       number of observations in at least one regime is less than the
c            required minimum mspan: IFAULT=-4 
  
c +++    decoding chromosome ends

      IF(IFAULT .NE.  0)  RETURN

c +++    long autoregression starts

      do 4 j =1 , nhb
4     res(j) = 0.0D-000

c       finds the residuals by means of a long autoregression for regime
c        Hannan Rissanen - the AR order is found by min AIC / BIC

c      write(lprint, 153) nkp
c153   format(1h  ,'Long autoregression, p = ', i6)

      do  11 ii  =1  ,  k

      nhmaxp =  nhp0	!=1
      hhcrit  = big

      do  1  ikj =  nhp0 , nhp1

      nkp  =  ikj
      nkp1 =  nkp +  1
      if(ii.le. 1)  then
      rlow =  -big
      else
      rlow =  r(ii - 1)
      endif
      rhigh =  r(ii)

c        selects the observations that belongs to the regime ii
c     and builds up the regression matrix for the long autoregression

      nkb  =  0
      nm1 =  nkp + 1
      do  23  j = nm1 ,  nhb
      if(z(j - nd) .gt. rlow .and. z(j - nd) .le. rhigh)  then
      nkb =  nkb  + 1
      x(nkb) = z(j)
      kkkn(nkb) =  j
      xmat(nkb, 1) = 1D0
      IF(NKP .LE. 0)  GOTO  55
      do  5  lagh =  1 ,  nkp
      lag1  =  lagh + 1
      jj  =  j - lagh
      if(jj  .ge.  1)  then
      xmat(nkb , lag1)  =  z(jj)
      else
      xmat(nkb, lag1)  =  0.0D-000 
      endif
5     continue 
55    CONTINUE
      endif
23    continue
      CALL REGRES(x, XMAT, mob, nkb, PHI, nkp1, AUX1, VC1, mx, BB, CC
     +, IP, IQ, TOL, MSING)

c       long autoregression estimate, singular matrix IFAULT = -5

      IF(MSING  .EQ.  2)  THEN
      IFAULT  =  -5
      RETURN
      ENDIF

c        residuals calculation with time point recovery

c      WRITE(LPRINT, 123)  II
c123   FORMAT(1H  ,'LONG AUTOREGRESSION: X,VFIT,RES. REGIME ',  I5)

      sskp =  0D0
      do  9  jj =  1  ,  nkb !conta le osservazioni de regime
      zo1 =  0D0
      do  10  ij  =  1 , nkp1 !ordine AR (varia da 1 a nhpo)
10    zo1 = zo1 +  phi(ij) * xmat(jj, ij)
      vfit(jj) =  zo1 
      sskp =  sskp + (x(jj) - vfit(jj)) * (x(jj) - vfit(jj))
9     continue
   
C +++++  end loop for choosing the "long" AR order
C         computing statistics AIC / BIC

      if(icri  .eq.   1)  then
      hkrite =  dble(nkb) * dlog(dabs(sskp) / dble(nkb))
     + + dble(nkp1) * 2D0
      else
      hkrite = dble(nkb) * dlog(dabs(sskp) / dble(nkb))
     +  + dble(nkp1) * dlog(dble(nkb)) 
      endif



c	WRITE(LPRINT, 1222) HKRITE
c1222  FORMAT(1H  ,'aic=', F10.5)

      if(hkrite .lt. hhcrit)  then
      hhcrit =  hkrite
      nhmaxp  =  nkp !ordine dell'autoregressivo lungo

C     IN QUESTO PUNTO HO SCELTO L'ORDINE DELL'AR LUNGO NHMAXP
C ORA COLCOLO LE INNOVAZIONI

C CALCOLO NKB INNOVAZIONI, SOLO SE .....
      do  2  jj =  1 , nkb !NKB=OSS DEL REGIME
      j =  kkkn(jj)
      res(j)  =  x(jj) - vfit(jj)


c      WRITE(LPRINT, 124) JJ, J, X(JJ), VFIT(JJ), RES(J)
c124   FORMAT(1H  ,2I5,  3F10.4)

2     continue
c      WRITE(LPRINT, 124) NKB
c124   FORMAT(1H  ,'....->', i5)

      endif
C	WRITE(LPRINT, 124) NKB
C124   FORMAT(1H  ,'....->', i5)

c      WRITE(LPRINT, 123)  II
c123   FORMAT(1H  ,'..... Regime ',  I5)
c      SSV  =  SSKP / DBLE(NKB)
c      WRITE(LPRINT, 125) SSKP, SSV
c125   FORMAT(1H  ,'RESIDUAL SSQ, VAR ', 2E16.5)
c      WRITE(LPRINT, 139)
c139   FORMAT(1H   ,'PARAMETER ESTIMATES')
c      WRITE(LPRINT,  138) (PHI(IJ), IJ=1, NKP1)
c138   FORMAT(1H   ,7F10.4)

1     continue

      nhp(ii)  =  nhmaxp !nhp racchiude l'ordine dell'AR lungo

c	 WRITE(LPRINT, 1004) NHMAXP
c1004   FORMAT(1H  ,'NHMAXP', i5)

11    CONTINUE

c fino a questo punto ho calcolato AR lungo per ogni regime

c +++    computing long autoregression ends
c        residuals are available res(1)..res(nhb)

C      WRITE(LPRINT,  126)
C126   FORMAT(1H  ,'LONG AUTOREGRESSION: Z,RES. ALL REGIMES')
C      DO 127  J =1 , NHB
C127   WRITE(LPRINT, 128)  J, Z(J), RES(J)
C128   FORMAT(1H  ,I5, 2F10.4)  
C      WRITE(LPRINT,  155)
C155   FORMAT(1H  ,'SUBSET ARMA.....')

c +++    subset ARMA computation starts

      DO  419  IIII = 1 , 5
      DO  419   JJJJ = 1, 13
419   PPP(IIII, JJJJ) =  0.0

      do  12  ii = 1  ,  k
      nxxp =  nxp(ii)	   !ordine dell'AR
      nxxq =  nxq(ii)	   !ordine dell'MA
      nkp =  nxxp + nxxq !Numero totale di parametri
      nkp1  =  nkp + 1
      if(ii.le. 1)  then
      rlow =  -big
      else
      rlow =  r(ii - 1)
      endif
      rhigh =  r(ii)

c    Re-builds the matrix XMAT for AR and MA subset parameter estimate
c    It is not necessary to set some integer n0=max(nd,np), as in Tong
c    and Lim, 1980, p. 257, because the Hannan and Rissanen method needs
c    M observations for the long autoregression, and, in addition, the
c    subset ARMA requires max(P,Q) lags to be discarded, so that nd<np
     
      nkb  =  0
c	nm1 =  nhp(ii) + MAX0(NPLAG , NQLAG) + 1
 
c      nm1 =  nhp(ii) + max0(nxxp, nxxq) + 1 !oss che scarto
       nm1=nhp1+ MAX0(NPLAG , NQLAG) + 1
C	nm1=	 Max0(nhp1,MAX0(NPLAG , NQLAG) + 1 )
      do  6  j = nm1 ,  nhb  !FONDAMENTALE
      if(z(j - nd) .gt. rlow .and. z(j - nd) .le. rhigh)  then
      nkb =  nkb  + 1
      x(nkb) = z(j)
      kkkn(nkb)  = j
      xmat(nkb, 1) = 1D0
      IF(NXXP  .LE.  0) GOTO  77
      do  7  lagh =  1 ,  nxxp
      ntp  =  lahp(ii , lagh)
      lag1  =  lagh + 1
      jj  =  j - ntp
      if(jj  .ge.  1)  then
      xmat(nkb , lag1)  =  z(jj)
      else
      xmat(nkb, lag1)  =  0.0D-000 
      endif
7     continue 
77    CONTINUE
      IF(NXXQ  .LE.  0)  GOTO  88
      do  8  lagh =  1 ,  nxxq
      ntq  =  lahq(ii , lagh)
      lag1  =  NXXP + lagh + 1
      jj  =  j - ntq
      if(jj  .ge.  1)  then
      xmat(nkb , lag1)  =  res(JJ)
      else
      xmat(nkb, lag1)  =  0.0D-000 
      endif
8     continue 
88    CONTINUE
      endif
6     continue
      nkp =  nxxp + nxxq
      nkp1  =  nkp + 1

C      WRITE(LPRINT, 1414) NKB
C1414   FORMAT(1H  ,'NHB PER ARMA', i5) 

      CALL REGRES(x, XMAT, mob, nkb, PHI, nkp1, AUX1, VC1, mx, BB, CC
     +, IP, IQ, TOL, MSING)

c       subset ARMA estimate, singular matrix IFAULT = -6

      IF(MSING  .EQ.  2)  THEN
      IFAULT  =  -6
      RETURN
      ENDIF

c        residuals and residual sum of squares calculation 

c      WRITE(LPRINT, 129)  II
c129   FORMAT(1H  ,'SUBSET ARMA: X,VFIT,RES. REGIME ',  I5)

      sskp =  0D0
      do  19  jj =  1  ,  nkb
      zo1 =  0D0
      do  18  ij  =  1 , nkp1
18    zo1 = zo1 +  phi(ij) * xmat(jj, ij)
      vfit(jj) =  zo1 
      sskp =  sskp + (x(jj) - vfit(jj)) * (x(jj) - vfit(jj))
      j =  kkkn(jj)
      res(j)  =  x(jj) - vfit(jj)	 !CALCOLO NkB RESIDUI

c      WRITE(LPRINT, 130) JJ, J, X(JJ), VFIT(JJ), RES(J)
c130   FORMAT(1H  ,2I5,  3F10.4)

19    continue

      if(iprint .eq. 1)  then
      WRITE(LPRINT, 151) II, NKB
151   FORMAT(1H    ,'....Regime ', I6,5x, 'NHB PER ARMA', i5)
      SSV  =  SSKP / DBLE(NKB)
      WRITE(LPRINT, 131) SSKP, SSV
131   FORMAT(1H  ,'RESIDUAL SSQ, VAR ', 2E16.5)
      WRITE(LPRINT, 140)
140   FORMAT(1H   ,'PARAMETER ESTIMATES')
      WRITE(LPRINT,  137) (PHI(IJ), IJ=1, NKP1)
137   FORMAT(1H   ,7F10.4)
      endif

      DO 417 IJ=1, NKP1
417   PPP(II,IJ) = PHI(IJ) !PARAMETRI PER REGIME
      nj(ii)  =  nkb !NUMERO DI OSSERVAZIONI IN OGNI REGIME
      s2(ii)  =  sskp !SOMMA DEI QUADRATI

c +++     subset ARMA computation ends

12    continue            


c      WRITE(LPRINT,  132)
c132   FORMAT(1H  ,'SUBSET ARMA: Z,RES. ALL REGIMES')
c      DO 133  J =1 , NHB
c133   WRITE(LPRINT, 134)  J, Z(J), RES(J)
c134   FORMAT(1H  ,I5, 2F10.4)  

c +++     AIC / BIC criterion computation and fitness function starts

      crki =  0.0D-000
      do 40 ii  =1  ,  k
      sskp  =  s2(ii)
      ntpq =  nxp(ii) + nxq(ii) + 1 !NUMERO TOTALE PARAMETRI
      ik  =  nj(ii)!NUMERO DI OSSERVAZIONI PER REGIME

c Formula (8.6) in Tong and Lim (1980) [cfr. anche Wong and Li, 1998
c p. 119] d= delay parameter; L= max order for each of the AR models
c n0= max(d,L); L e' riferito alla "long autoregression" piu' il modello
c ARMA. La normalizzazione dell'AIC crki/(nhb-n0) serve con

      

      if(icri  .eq.  1)  then
      chrki = DBLE(IK) * dlog(sskp / dble(ik)) + dble(ntpq) * 2D0 
      else
      chrki = DBLE(IK)*dlog(sskp/dble(ik)) + dble(ntpq)*dlog(dble(ik))
      endif      
      crki  =  crki +  chrki    

c      WRITE(LPRINT, 135) II
c135   FORMAT(1H   ,'REGIME ', I5)
c      WRITE(LPRINT,  136) SSKP, NJ(II), NTPQ, IK, CHRKI
c136   FORMAT(1H   ,'SSKP,NJ,NTPQ,IK,CHRKI'/1H   ,E12.4, 3I10, E12.4) 

40    CONTINUE
c	n0 =  max0(nd, max0(nhp1, max0(nplag, nqlag)))
	n0 =  max0(nd, NM1-1)

	crki= crki / dble(nhb-n0)
      FFB  =  dexp(-crki / c)

c +++     AIC / BIC criterion computation and fitness function ends

      if(iprint  .eq.  1)  then
      WRITE(LPRINT, 100) nchro
100   FORMAT(1H  ,'chromosome vector LCHROM length = ', I5)
      WRITE(LPRINT, 101) (LCHROM(II), II=1,  NCHRO)
101   FORMAT(1H   ,  70I1)
      WRITE(LPRINT, 102) nd
102   FORMAT(1H  ,'delay parameter ND = ', I5)
      WRITE(LPRINT, 103) k
103   FORMAT(1H  ,'number of regimes K = ', I5)
      WRITE(LPRINT, 104)
104   FORMAT(1H  ,'threshold vector R')
      WRITE(LPRINT, 105) (R(IK), IK=1,  k1)
105   format(1h  ,10f7.4)
      WRITE(LPRINT, 1015) (NHP(II), ii=1,  k)
1015   format(1h  ,'Ordine AR Lungo', I5)
      WRITE(LPRINT, 106)
106   FORMAT(1H  ,'nobs per regime vector NJ')
      WRITE(LPRINT, 107) (NJ(IK), IK=1,  k)
107   FORMAT(1H   ,  14I5)
      WRITE(LPRINT, 108) 
108   FORMAT(1H   ,  'sum of squares per regime S2')
      write(lprint , 109)  (s2(IK), IK=1,  k)
109   format(1h  , 7F10.4)
      write(lprint, 110)
110   format(1h  ,'ARMA subset for each regime')
      sskp  =  0.0D-000
      do  111  ii  =1  ,  k
      write(lprint, 112) ii
112   format(1h   ,'regime n. ', I5)
      nxxp = nxp(ii)
      write(lprint, 113) nxxp
113   format(1h  ,'AR order nxxp = ', I5)
      write(lprint, 114) 
114   format(1h  ,'subset AR lags')
      write(lprint, 115) (lahp(ii, j), j=1, nxxp)
115   format(1h   ,14I5)
      nxxq = nxq(ii)
      write(lprint, 116) nxxq
116   format(1h  ,'MA order nxxq = ', I5)
      write(lprint, 117) 
117   format(1h  ,'subset MA lags')
      write(lprint, 118) (lahq(ii, j), j=1, nxxq)
118   format(1h   ,14I5)
      sskp =  sskp + s2(ii)
111   continue
      nosser=0.0
      do 1980 i= 1, k
1980	nosser= nosser + nj(i)
      ssv  =  sskp / dble( NOSSER)
      write(lprint, 119) sskp, ssv
119 	format(1h  ,'pooled ssq and var =  ', 2E16.5)
      write(lprint, 120) icri, crki
120   format(1h  ,'criterion, NAIC value ', I5, E16.5)
      AICN = crki * (dble(nhb - n0)) 
	write(lprint, 1720) AICN
1720   format(1h  ,'NAIC/Effettive Osservazioni  ', E16.5)
     
      write(lprint, 121) c 
121   format(1h  ,'adjustment constant c (xeta) ',  E16.5)
      write(lprint, 122) ffb
122   format(1h   ,'Exit FFB = ', D14.6)
      endif

      RETURN
      END


      SUBROUTINE DECODE(LCHROM, MCHRO, NCHRO, Z, NM, NHB, KKHN, ND, R
     +,MK, NK, NP, NQ, NXP, LAHP, NXQ, LAHQ, NBIT, NBIT1, NBIT2, MSPAN
     +, K, K1, nplag, nqlag, IFAULT)

C       decoding the variable-length chromosome that encodes a SETARMA
C       subset model: ND delay parameter, K regimes, NXP and NXQ orders
C       LAHP , LAHQ autoregressive and moving average lags respectively

      IMPLICIT REAL*8 (A - H , O - Z)
      DIMENSION LCHROM(MCHRO), z(nhb), kkhn(nhb), r(nk)
     +, nxp(nk), lahp(MK, nplag), nxq(nk), lahq(MK, nqlag)
      data  tol/ 8.43D-37/
       data     big  / 3.37D+38 /                                       

      lprint  =6

c      write(lprint, 100)
c100   format(1h  ,'Enter DECODE ')
c      write(lprint, 101) nbit
c101   format(1h   ,'Integer encoding nbit = ', I5)
c      write(lprint, 102)  mchro
c102   format(1h  ,'mchro (max chromosome length) = ', I8)
c      write(lprint, 103) (lchrom(i), i=1, mchro)
c103   format(1h  ,70I1)

c        initialize the chromosome LCHROM bit count ik 

      IK  =  0

c        decoding the delay parameter ND

      nd =  0
      do   20   ii =  1 ,  nbit
      ik  =  ik  + 1
20    nd =  nd + lchrom(ik) * (2 ** (nbit - ii))

c      write(lprint, 104) nd
c104   format(1h  ,'delay parameter nd = ', I5)

      if(nd .le. 0)  then
      ifault = -7
      return
      endif

c        decoding the number of regimes and the threshold parameters
c        number of threshold is k1, number of regimes is k = k1 + 1

      k1  =  0
      n1  =  mspan + 1
      n2  =  nhb -  mspan
      DO  1   II  =  n1 , n2
      IK = IK +  1

c      WRITE(LPRINT, 200) II, IK, LCHROM(IK)
c200   FORMAT(1H   ,3I5)

      IF(LCHROM(Ik) .GE. 1) THEN
      k1 =  k1 + 1

c       number of regimes greater than the maximum allowed nk: IFAULT=-1 

       if(k1 .gt. nk - 1) then
       ifault =  -1
       return
       endif

      JJ = KKHN(II)
      R(k1)  =  Z(JJ) 
      ENDIF
1     CONTINUE
      K  =  k1  +  1
      R(k)  =   big

c      write(lprint, 105) k1
c105   format(1h   ,'k1 = ', i5)
c      write(lprint, 106)
c106   format(1h   ,'r(1)..r(k)')
c      write(lprint, 107) (r(i), i=1, k)
c107   format(1h   ,6E12.4) 

c      for each regime, checks if the observed values are at least mspan

c      write(lprint, 108) nm
c108   format(1h  ,'nm = ', I5)

      nm1 =  nm  + 1
      do  221  ii  =1  ,  k

c      write(lprint, 109) ii
c109   format(1h  ,'ii = ',  I5)

      if(ii.le. 1)  then
      rlow =  -big
      else
      rlow =  r(ii - 1)
      endif
      rhigh =  r(ii)
      nkb  =  0
      do  222  j = nm1 ,  nhb
      if(z(j - nd) .gt. rlow .and. z(j - nd) .le. rhigh)  then
      nkb =  nkb  + 1
      endif
222   continue
      
c       number of observations in the current regime II is less
c          than the required minimum mspan: IFAULT=-4
 
c      write(lprint, 110) nkb
c110   format(1h  ,'nkb = ', I5)

      IF(nkb  .lt.  mspan)  THEN
      IFAULT  =  -4
      RETURN
      ENDIF

221   continue

c       for each regime, decodes the identified subset ARMA model

      do  2  ii =1  ,  k

c         decodes the autoregressive  order p

      nxxp  =  0
      do  30  jj =  1,  nbit1
      ik  =  ik + 1
30    nxxp =  nxxp + lchrom(ik) * (2 ** (nbit1 - jj))
       nxxp=min0(np,max0(0,idnint(dble(nxxp)*dble(np)/(2**nbit1-1))))
c       the maximum AR order is greater than np: IFAULT=-2

       if(nxxp .gt. np) then
       ifault =  -2
       return
       endif

      nxp(ii)  = nxxp

c      write(lprint, 111) ii,  nxxp
c111   format(1h    ,'ii,nxxp = ',2I5) 

c         decodes subset autoregressive lag

      if(nxxp  .le.  0)  goto  220
      do  1220  ilag = 1, nxxp
      lagh  =  0
      do  330  jj =  1,  nbit2
      ik  =  ik + 1
330   lagh =  lagh + lchrom(ik) * (2 ** (nbit2 - jj))
       lagh
     +=min0(nplag,max0(1,idnint(dble(lagh)*dble(nplag)/(2**nbit2-1))))
c      write(lprint, 112) lagh
c112   format(1h   ,'lagh = ',  I5)

c       the maximum AR lag is greater than np: IFAULT=-2

       if(lagh .gt. nplag) then
       ifault =  -2
       return
       endif

c       the  AR lag is zero: IFAULT=-8

       if(lagh .le.  0) then
       ifault =  -8
       return
       endif

      lahp(ii , ilag) =  lagh

1220  continue
220   continue

c         decodes the moving average  order q

      nxxq  =  0
      do  31  jj =  1,  nbit1
      ik  =  ik + 1
31    nxxq =  nxxq + lchrom(ik) * (2 ** (nbit1 - jj))
       nxxq=min0(nq,max0(0,idnint(dble(nxxq)*dble(nq)/(2**nbit1-1))))
c       the maximum MA order is greater than nq: IFAULT=-3

       if(nxxq .gt. nq) then
       ifault =  -3
       return
       endif

      nxq(ii)  = nxxq

c      write(lprint, 113) ii,  nxxq
c113   format(1h    ,'ii,nxxq = ',2I5) 

c         decodes subset moving average lag

      if(nxxq  .le.  0)   goto  322
      do  22  ilag = 1, nxxq
      lagh  =  0
      do  32  jj =  1,  nbit2
      ik  =  ik + 1
32    lagh =  lagh + lchrom(ik) * (2 ** (nbit2 - jj))
       lagh
     +=min0(nqlag,max0(1,idnint(dble(lagh)*dble(nqlag)/(2**nbit2-1))))
c      write(lprint, 114) lagh
c114   format(1h   ,'lagh = ',  I5)

c       the maximum MA lag is greater than nq: IFAULT=-3

       if(lagh .gt. nqlag) then
       ifault =  -3
       return
       endif

c       the  MA lag is zero: IFAULT=-9

       if(lagh .le.  0) then
       ifault =  -9
       return
       endif

      lahq(ii , ilag) =  lagh

22    continue
322   continue

2     continue

      NCHRO = IK
      ifault  = 0
      return
      end


      SUBROUTINE REGRES(Y, XMAT, MOB, NOB, P, NP, A, D, MP, B, C, IP, IQ
     +, EPS, MSING)

C   Linear regression ordinary least squares : p = (X'X)^(-1) X'y
C   msing = 1, normal exit; msing = 2, the matrix X'X is singular

      IMPLICIT REAL*8 (A - H , O - Z)
      DIMENSION  Y(NOB), XMAT(MOB, NP), P(NP), A(NP) 
     +, D(MP, NP), B(NP), C(NP), IP(NP), IQ(NP)

C   X'X -> D (here X == XMAT)

      DO  1  I  =1  ,  NP
      DO 1  J  =1  ,  NP
      SUM  =  0.0
      DO  2  K  =1  ,  NOB
2     SUM  =  SUM  +  XMAT(K , I) * XMAT(K , J)
1     D(I , J)  =  SUM

C   (X'X)^(-1) -> D

      CALL  DGJR(D, NP, MP, EPS, MSING, B, C, IP, IQ)
      IF(MSING .EQ. 2)  RETURN

C   X'y -> a

      DO  3  I  =1  ,  NP
      SUM  =  0.0
      DO  4  K  =1  ,  NOB
4     SUM =  SUM + XMAT(K , I) * Y(K)
3     A(I)  =  SUM

C   (X'X)^(-1) X'y == Da -> p

      DO  5  I =1  ,  NP
      SUM  =  0.0
      DO  6  J  =1  ,  NP
6     SUM  =  SUM +  D(I , J) * A(J)
5     P(I) =  SUM
C                                                             End REGRES
      RETURN
      END


      SUBROUTINE DGJR(A,N,MM,EPS,MSING,B,C,P,Q)
C----     GAUSS-JORDAN-RUTISHAUSER MATRIX INVERSION WITH DOUBLE PIVOTING
      IMPLICIT REAL*8 (A - H , O - Z)
      INTEGER P,Q
      DIMENSION A(MM,N),B(N),C(N),P(N),Q(N)
      MSING=1
      DO 10 K=1,N
C----     DETERMINATION OF THE PIVOT ELEMENT
      PIVOT=0.0
      DO 20 I=K,N
      DO 20 J=K,N
      IF(DABS(A(I,J))-DABS(PIVOT)) 20,20,30
30    PIVOT=A(I,J)
      P(K)=I
      Q(K)=J
20    CONTINUE
      IF(DABS(PIVOT)-EPS) 40,40,50
C----     EXCHANGE OF THE PIVOTAL ROW WITH THE K-TH ROW
50    IF(P(K)-K) 60,80,60
60    DO 70 J=1,N
      L=P(K)
      Z=A(L,J)
      A(L,J)=A(K,J)
70    A(K,J)=Z
C----     EXCHANGE OF THE PIVOTAL COLUMN WITH THE K-TH COLUMN
80    IF(Q(K)-K) 85,90,85
85    DO 100 I=1,N
      L=Q(K)
      Z=A(I,L)
      A(I,L)=A(I,K)
100   A(I,K)=Z
90    CONTINUE
C----     JORDAN STEP
      DO 110 J=1,N
      IF(J-K) 130,120,130
120   B(J)=1.0/PIVOT
      C(J)=1.0
      GOTO 140
130   B(J)=-A(K,J)/PIVOT
      C(J)=A(J,K)
140   A(K,J)=0.0
110   A(J,K)=0.0
      DO 10 I=1,N
      DO 10 J=1,N
10    A(I,J)=A(I,J)+C(I)*B(J)
C----     REORDERING THE MATRIX
      DO 155 M=1,N
      K=N-M+1
      IF(P(K)-K) 160,170,160
160   DO 180 I=1,N
      L=P(K)
      Z=A(I,L)
      A(I,L)=A(I,K)
180   A(I,K)=Z
170   IF(Q(K)-K) 190,155,190
190   DO 150 J=1,N
      L=Q(K)
      Z=A(L,J)
      A(L,J)=A(K,J)
150   A(K,J)=Z
155   CONTINUE
151   RETURN
40    CONTINUE
      MSING=2
      GOTO 151
      END
                                                                       

      REAL*8 FUNCTION DOTPRD(P, X, Y)                                   ARI03050
C                                                                       ARI03060
C  ***  RETURN THE INNER PRODUCT OF THE P-VECTORS X AND Y.  ***         ARI03070
C                                                                       ARI03080
C                                                                       ARI03090
C     THIS PROGRAM IS TAKEN FROM :                                      ARI03100
C     J.E.DENNIS, D.M.GAY AND R.E.WELSCH, 'ALGORITHM 573 NL2SOL-        ARI03110
C     AN ADAPTIVE NONLINEAR LEAST-SQUARES ALGORITHM E4| ',             ARI03120
C     ACM TRANSECTIONS ON MATHEMATICAL SOFTWARE, VOL.7, NO.3,           ARI03130
C     SEPTEMBER 1981                                                    ARI03140
C                                                                       ARI03150
      INTEGER P                                                         ARI03160
      REAL*8 X(P), Y(P)                                                 ARI03170
C                                                                       ARI03180
      INTEGER I                                                         ARI03190
      REAL ONE, SQTETA, T, ZERO                                         ARI03200
C/+                                                                     ARI03210
      REAL*8 DMAX1, DABS                                                ARI03220
      REAL  ABS
C/                                                                      ARI03230
      EXTERNAL RMDCON                                                   ARI03240
      REAL RMDCON                                                       ARI03250
C                                                                       ARI03260
C  ***  RMDCON(2) RETURNS A MACHINE-DEPENDENT CONSTANT, SQTETA, WHICH   ARI03270
C  ***  IS SLIGHTLY LARGER THAN THE SMALLEST POSITIVE NUMBER THAT       ARI03280
C  ***  CAN BE SQUARED WITHOUT UNDERFLOWING.                            ARI03290
C                                                                       ARI03300
C/6                                                                     ARI03310
      DATA ONE/1.E+0/, SQTETA/0.E+0/, ZERO/0.E+0/                       ARI03320
C/7                                                                     ARI03330
C     PARAMETER (ONE=1.D+0, ZERO=0.D+0)                                 ARI03340
C     DATA SQTETA/0.D+0/                                                ARI03350
C/                                                                      ARI03360
C                                                                       ARI03370
      DOTPRD = ZERO                                                     ARI03380
      IF (P .LE. 0) GO TO 999                                           ARI03390
      IF (SQTETA .EQ. ZERO) SQTETA = RMDCON(2)                          ARI03400
      DO 20 I = 1, P                                                    ARI03410
         T = DMAX1(DABS(X(I)), DABS(Y(I)))                              ARI03420
         IF (T .GT. ONE) GO TO 10                                       ARI03430
         IF (T .LT. SQTETA) GO TO 20                                    ARI03440
         T = (X(I)/SQTETA)*Y(I)                                         ARI03450
         IF (ABS(T) .LT. SQTETA) GO TO 20                               ARI03460
 10      DOTPRD = DOTPRD + X(I)*Y(I)                                    ARI03470
 20   CONTINUE                                                          ARI03480
C                                                                       ARI03490
 999  RETURN                                                            ARI03500
C  ***  LAST CARD OF DOTPRD FOLLOWS  ***                                ARI03510
      END                                                               ARI03520


      REAL FUNCTION RMDCON(K)                                           ARI14080
                                                                        ARI14090
C  ***  RETURN MACHINE DEPENDENT CONSTANTS                 ***          ARI14100
C                                                                       ARI14110
C     THIS PROGRAM IS TAKEN FROM :                                      ARI14120
C     J.E.DENNIS, D.M.GAY AND R.E.WELSCH, 'ALGORITHM 573 NL2SOL-        ARI14130
C     AN ADAPTIVE NONLINEAR LEAST-SQUARES ALGORITHM E4| ',             ARI14140
C     ACM TRANSECTIONS ON MATHEMATICAL SOFTWARE, VOL.7, NO.3,           ARI14150
C     SEPTEMBER 1981                                                    ARI14160
C                                                                       ARI14170
C +++  COMMENTS BELOW CONTAIN DATA STATEMENTS FOR VARIOUS MACHINES.  +++ARI14180
C +++  TO CONVERT TO ANOTHER MACHINE, PLACE A C IN COLUMN 1 OF THE   +++ARI14190
C +++  DATA STATEMENT LINE(S) THAT CORRESPOND TO THE CURRENT MACHINE +++ARI14200
C +++  AND REMOVE THE C FROM COLUMN 1 OF THE DATA STATEMENT LINE(S)  +++ARI14210
C +++  THAT CORRESPOND TO THE NEW MACHINE.                           +++ARI14220
C                                                                       ARI14230
      INTEGER K                                                         ARI14240
C                                                                       ARI14250
C  ***  THE CONSTANT RETURNED DEPENDS ON K...                           ARI14260
C                                                                       ARI14270
C  ***        K = 1... SMALLEST POS. ETA SUCH THAT -ETA EXISTS.         ARI14280
C  ***        K = 2... SQUARE ROOT OF 1.001*ETA.                        ARI14290
C  ***        K = 3... UNIT ROUNDOFF = SMALLEST POS. NO. MACHEP SUCH    ARI14300
C  ***                 THAT 1 + MACHEP .GT. 1 .AND. 1 - MACHEP .LT. 1.  ARI14310
C  ***        K = 4... SQUARE ROOT OF 0.999*MACHEP.                     ARI14320
C  ***        K = 5... SQUARE ROOT OF 0.999*BIG (SEE K = 6).            ARI14330
C  ***        K = 6... LARGEST MACHINE NO. BIG SUCH THAT -BIG EXISTS.   ARI14340
C                                                                       ARI14350
      REAL BIG, ETA, MACHEP                                             ARI14360
                                                                        ARI14370
      REAL SQRT                                                         ARI14380
                                                                        ARI14390
      REAL ONE001, PT999                                                ARI14400
                                                                        ARI14410
      DATA ONE001/1.001/, PT999/0.999/                                  ARI14420
                                                                        ARI14430
C  +++  IBM 360, IBM 370, OR XEROX  +++                                 ARI14440
C                                                                       ARI14450
C      DATA BIG/Z7FFFFFFF/, ETA/Z00100000/, MACHEP/Z3C100000/           ARI14460
C                                                                       ARI14470
C  +++  DATA GENERAL  +++                                               ARI14480
C                                                                       ARI14490
C     DATA BIG/0.7237E+76/, ETA/0.5398E-78/, MACHEP/0.9537E-06/         ARI14500
C                                                                       ARI14510
C  +++  DEC 11  +++                                                     ARI14520
C                                                                       ARI14530
C     DATA BIG/1.7E+38/, ETA/2.9388E-39/, MACHEP/1.1921E-07/            ARI14540
C                                                                       ARI14550
C  +++  HP3000  +++                                                     ARI14560
C                                                                       ARI14570
C     DATA BIG/1.1579E+77/, ETA/8.6362E-78/, MACHEP/2.3842E-07/         ARI14580
C                                                                       ARI14590
C  +++  HONEYWELL  +++                                                  ARI14600
C                                                                       ARI14610
C     DATA BIG/O376777000000/, ETA/O404400400000/,                      ARI14620
C    1     MACHEP/O716400000000/                                        ARI14630
C                                                                       ARI14640
C  +++  DEC10  +++                                                      ARI14650
C                                                                       ARI14660
C     DATA BIG/"377777777777/, ETA/"000400000021/,                      ARI14670
C    1     MACHEP/"147400000000/                                        ARI14680
C                                                                       ARI14690
C  +++  BURROUGHS  +++                                                  ARI14700
C                                                                       ARI14710
C     DATA BIG/O0777777777777777/, ETA/O1771000000000000/,              ARI14720
C    1     MACHEP/O1301000000000000/                                    ARI14730
C                                                                       ARI14740
C  +++  CONTROL DATA  +++                                               ARI14750
C                                                                       ARI14760
C     DATA BIG/37754000000000000000B/, ETA/00024000000000000000B/,      ARI14770
C    1     MACHEP/16414000000000000000B/                                ARI14780
C                                                                       ARI14790
C  +++  PRIME  +++                                                      ARI14800
C                                                                       ARI14810
C     DATA BIG/1.7E+38/, ETA/1.47E-39/, MACHEP/2.38419E-7/              ARI14820
C                                                                       ARI14830
C  +++  UNIVAC  +++                                                     ARI14840
C                                                                       ARI14850
C     DATA BIG/1.69E+38/, ETA/5.9E-39/, MACHEP/1.4901162E-8/            ARI14860
C                                                                       ARI14870
C  +++  VAX  +++                                                        ARI14880
C                                                                       ARI14890
C     DATA BIG/1.7E+38/, ETA/2.939E-39/, MACHEP/5.9604645E-08/          ARI14900
C
c  approximations of real numbers in floating-point form (1 to 7 digits)        
c  from 8.43E-37 to 3.37E+38 (assumed zero if less/equal than 0.000015)
        big  =  3.37E+38
        eta  =   8.43E-37 
        machep  = 0.000015 
C                                                                       ARI14910
C-------------------------------  BODY  --------------------------------ARI14920
                                                                        ARI14930
      GO TO (10, 20, 30, 40, 50, 60), K                                 ARI14940
                                                                        ARI14950
 10   RMDCON = ETA                                                      ARI14960
      GO TO 999                                                         ARI14970
                                                                        ARI14980
 20   RMDCON = SQRT(ONE001*ETA)                                         ARI14990
      GO TO 999                                                         ARI15000
                                                                        ARI15010
 30   RMDCON = MACHEP                                                   ARI15020
      GO TO 999                                                         ARI15030
                                                                        ARI15040
 40   RMDCON = SQRT(PT999*MACHEP)                                       ARI15050
      GO TO 999                                                         ARI15060
                                                                        ARI15070
 50   RMDCON = SQRT(PT999*BIG)                                          ARI15080
      GO TO 999                                                         ARI15090
                                                                        ARI15100
 60   RMDCON = BIG                                                      ARI15110
                                                                        ARI15120
 999  RETURN                                                            ARI15130
C  ***  LAST CARD OF RMDCON FOLLOWS  ***                                ARI15140
      END                                                               ARI15150
                                                                        ARI15160

      REAL*8 FUNCTION SUMF(X,N)
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION X(N)                                                    
      SUMF=0.0                                                          
      DO 10 I=1,N                                                       
   10 SUMF=SUMF+X(I)                                                    
      RETURN                                                            
      END                                                               


      REAL*8 FUNCTION RMEAN(X,NOB)                                      ARI01510
C                                                                       ARI01520
C----      DOUBLE PRECISION FUNCTION RMEAN                              ARI01530
C                                                                       ARI01540
C      CALCULATES THE SAMPLE MEAN OF A SERIES OF NOB OBSERVATIONS       ARI01550
C                                                                       ARI01560
C----                                                                   ARI01570
C                                                                       ARI01580
C      THIS PROGRAM IS TAKEN FROM " A COMPUTER PROGRAM FOR THE          ARI01590
C      ANALYSIS OF TIME SERIES MODELS USING THE BOX-JENKINS PHILOSOPHY "ARI01600
C      BY D.J.PACK (1979) AUTOMATIC FORECASTING SYSTEMS, INC.           ARI01610
C                                                                       ARI01620
C----                                                                   ARI01630
C                                                                       ARI01640
C      X:  OBSERVED TIME SERIES OF NOB VALUES                           ARI01650
C                                                                       ARI01660
C----                                                                   ARI01670
C                                                                       ARI01680
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION X(NOB)                                                  ARI01690
      SUM=0                                                             ARI01700
      DO 10 J=1,NOB                                                     ARI01710
  10  SUM=SUM+X(J)                                                      ARI01720
      RMEAN=SUM/FLOAT(NOB)                                              ARI01730
      RETURN                                                            ARI01740
      END                                                               ARI01750


      REAL*8 FUNCTION CPROXY(X,XBAR,Y,YBAR,NOB)                         ARI01760
C                                                                       ARI01770
C----     DOUBLE PRECISION FUNCTION CPROXY                              ARI01780
C                                                                       ARI01790
C      CALCULATES SUM OF PRODUCTS OF DEVIATIONS OF TWO VARIABLES        ARI01800
C      FROM SPECIFIED CENTRAL VALUES                                    ARI01810
C                                                                       ARI01820
C----                                                                   ARI01830
C                                                                       ARI01840
C      THIS PROGRAM IS TAKEN FROM " A COMPUTER PROGRAM FOR THE          ARI01850
C      ANALYSIS OF TIME SERIES MODELS USING THE BOX-JENKINS PHILOSOPHY "ARI01860
C      BY D.J.PACK (1979) AUTOMATIC FORECASTING SYSTEMS, INC.           ARI01870
C                                                                       ARI01880
C----                                                                   ARI01890
C                                                                       ARI01900
C      X,Y:      REAL ARRAYS OF LENGTH NOB.                             ARI01910
C                ON ENTRY CONTAIN VALUES OF THE TWO VARIABLES           ARI01920
C      XBAR,YBAR:X'S AND Y'S MEAN RESPECTIVELY                          ARI01930
C                                                                       ARI01940
C----                                                                   ARI01950
C                                                                       ARI01960
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION X(NOB),Y(NOB)                                           ARI01970
      S=0                                                               ARI01980
      DO 10 I=1,NOB                                                     ARI01990
  10  S=S+(X(I)-XBAR)*(Y(I)-YBAR)                                       ARI02000
      CPROXY=S                                                          ARI02010
      RETURN                                                            ARI02020
      END                                                               ARI02030


      SUBROUTINE ZCOR(C,NOB,NAC,R,SER,DM,DSD,IEXIT)                     ARI02040
C                                                                       ARI02050
C-----     SUBROUTINE ZCOR                                              ARI02060
C                                                                       ARI02070
C     CALCULATES AUTOCORRELATIONS AND THEIR STANDARD ERRORS             ARI02080
C                                                                       ARI02090
C-----                                                                  ARI02100
C                                                                       ARI02110
C     C     SERIES INPUT OF NOB OBS.                                    ARI02120
C     NAC   MAXIMUM LAG                                                 ARI02130
C     R     AUTOCORRELATION FUNCTION OF NAC VALUES                      ARI02140
C     SER  AUTOCOVARIANCE FUNCTION OF NAC VALUES                        ARI02150
C     DM    MEAN OF THE SERIES                                          ARI02160
C     DSD   STANDARD ERROR OF THE SERIES                                ARI02170
C     IEXIT IF NAC>= NOB IEXIT IS SET TO -1                             ARI02180
C           UNLESS THIS ONE ERROR IS DETECTED IEXIT                     ARI02190
C           CONTAINS ZERO ON EXIT                                       ARI02200
C                                                                       ARI02210
C-----                                                                  ARI02220
C                                                                       ARI02230
      IMPLICIT REAL*8(A-H , O-Z)
      DIMENSION C(NOB),R(NAC),SER(NAC)                                  ARI02240
      IEXIT=0                                                           ARI02250
      IF(NOB-NAC) 100,100,20                                            ARI02260
   20 DIV=NOB                                                           ARI02270
      DM=RMEAN(C,NOB)                                                   ARI02280
      DENOM=CPROXY(C,DM,C,DM,NOB)                                       ARI02290
      DSD=DSQRT(DENOM/DIV)                                              ARI02300
      DO 30 K=1,NAC                                                     ARI02310
      M=NOB-K                                                           ARI02320
      HNUM=CPROXY(C(1),DM,C(K+1),DM,M)                                  ARI02330
      SER(K)=HNUM/DIV                                                   ARI02340
      R(K)=HNUM/DENOM                                                   ARI02350
   30 CONTINUE                                                          ARI02360
      NAC1=NAC+1                                                        ARI02370
      DO 40 K=1,NAC                                                     ARI02380
      M=NAC1-K+1                                                        ARI02390
      R(M)=R(M-1)                                                       ARI02400
      SER(M)=SER(M-1)                                                   ARI02410
   40 CONTINUE                                                          ARI02420
      R(1)=1.0                                                          ARI02430
      SER(1)=DSD*DSD                                                    ARI02440
      RETURN                                                            ARI02450
  100 IEXIT=-1                                                          ARI02460
      RETURN                                                            ARI02470
      END                                                               ARI02480


      SUBROUTINE VSUNIF(ISEME,P2)
      IMPLICIT INTEGER*4 (I-N)
      REAL*8 CC0,CC1,CC2,CC3,CC4,P2
      DATA CC2/16807.D0/,CC3/2147483647.D0/
      SAVE
      CC1=ISEME
      CC1=CC1*CC2
      CC4=DMOD(CC1,CC3)
      ISEME=CC4
      CC0=CC4
      P2=CC0/CC3
      RETURN
      END


      REAL*8 function  random(l)

c        algorithm as 183 Appl. Statist. (1982) Vol. 31, no. 2
c
c        returns a pseudo-random number rectangularly distributed 
c        between  0  and 1.
c
c        ix, iy and iz should be set to integer values between 
c        1 and 30000 before first entry
c
c        integer arithmetic up to 30323  is required

      IMPLICIT REAL*8 (A -H , O - Z)
	common /rand/  ix, iy,  iz
	ix = 171 * mod(ix, 177) - 2 * (ix / 177)
	iy = 172 * mod(iy, 176) - 35 * (iy / 176)
	iz = 170 * mod(iz, 178) - 63 * (iz / 178)

	if (ix .lt. 0) ix = ix + 30269
	if (iy .lt. 0) iy = iy + 30307
	if (iz .lt. 0) iz = iz + 30323

c        if integer arithmetic up to 5212632 is available, 
c        the preceding 6 statements may be replaced by
c
c        ix = mod(171 * ix, 30269)
c        iy = mod(172 * iy, 30307) 
c        iz = mod(170 * iz, 30323)
c
c        on some machines, this may slightly increase
c        the speed. the results will be identical.

	random  = DMOD(DBLE(ix) / 30269D0 + DBLE(iy) / 30307D0 +
     *                 DBLE(iz) / 30323D0, 1D0)

	return
	end


        DOUBLE PRECISION FUNCTION PPND16 (P, IFAULT)
C
C       ALGORITHM AS241  APPL. STATIST. (1988) VOL. 37, NO. 3
C
C       Produces the normal deviate Z corresponding to a given lower
C       tail area of P; Z is accurate to about 1 part in 10**16.
C
C       The hash sums below are the sums of the mantissas of the
C       coefficients.   They are included for use in checking
C       transcription.
C
        DOUBLE PRECISION ZERO, ONE, HALF, SPLIT1, SPLIT2, CONST1,
     *          CONST2, A0, A1, A2, A3, A4, A5, A6, A7, B1, B2, B3,
     *          B4, B5, B6, B7,
     *          C0, C1, C2, C3, C4, C5, C6, C7, D1, D2, D3, D4, D5,
     *          D6, D7, E0, E1, E2, E3, E4, E5, E6, E7, F1, F2, F3,
     *          F4, F5, F6, F7, P, Q, R
        PARAMETER (ZERO = 0.D0, ONE = 1.D0, HALF = 0.5D0,
     *          SPLIT1 = 0.425D0, SPLIT2 = 5.D0,
     *          CONST1 = 0.180625D0, CONST2 = 1.6D0)
C
C       Coefficients for P close to 0.5
C
        PARAMETER (A0 = 3.38713 28727 96366 6080D0,
     *             A1 = 1.33141 66789 17843 7745D+2,
     *             A2 = 1.97159 09503 06551 4427D+3,
     *             A3 = 1.37316 93765 50946 1125D+4,
     *             A4 = 4.59219 53931 54987 1457D+4,
     *             A5 = 6.72657 70927 00870 0853D+4,
     *             A6 = 3.34305 75583 58812 8105D+4,
     *             A7 = 2.50908 09287 30122 6727D+3,
     *             B1 = 4.23133 30701 60091 1252D+1,
     *             B2 = 6.87187 00749 20579 0830D+2,
     *             B3 = 5.39419 60214 24751 1077D+3,
     *             B4 = 2.12137 94301 58659 5867D+4,
     *             B5 = 3.93078 95800 09271 0610D+4,
     *             B6 = 2.87290 85735 72194 2674D+4,
     *             B7 = 5.22649 52788 52854 5610D+3)
C       HASH SUM AB    55.88319 28806 14901 4439
C
C       Coefficients for P not close to 0, 0.5 or 1.
C
        PARAMETER (C0 = 1.42343 71107 49683 57734D0,
     *             C1 = 4.63033 78461 56545 29590D0,
     *             C2 = 5.76949 72214 60691 40550D0,
     *             C3 = 3.64784 83247 63204 60504D0,
     *             C4 = 1.27045 82524 52368 38258D0,
     *             C5 = 2.41780 72517 74506 11770D-1,
     *             C6 = 2.27238 44989 26918 45833D-2,
     *             C7 = 7.74545 01427 83414 07640D-4,
     *             D1 = 2.05319 16266 37758 82187D0,
     *             D2 = 1.67638 48301 83803 84940D0,
     *             D3 = 6.89767 33498 51000 04550D-1,
     *             D4 = 1.48103 97642 74800 74590D-1,
     *             D5 = 1.51986 66563 61645 71966D-2,
     *             D6 = 5.47593 80849 95344 94600D-4,
     *             D7 = 1.05075 00716 44416 84324D-9)
C       HASH SUM CD    49.33206 50330 16102 89036
C
C       Coefficients for P near 0 or 1.
C
        PARAMETER (E0 = 6.65790 46435 01103 77720D0,
     *             E1 = 5.46378 49111 64114 36990D0,
     *             E2 = 1.78482 65399 17291 33580D0,
     *             E3 = 2.96560 57182 85048 91230D-1,
     *             E4 = 2.65321 89526 57612 30930D-2,
     *             E5 = 1.24266 09473 88078 43860D-3,
     *             E6 = 2.71155 55687 43487 57815D-5,
     *             E7 = 2.01033 43992 92288 13265D-7,
     *             F1 = 5.99832 20655 58879 37690D-1,
     *             F2 = 1.36929 88092 27358 05310D-1,
     *             F3 = 1.48753 61290 85061 48525D-2,
     *             F4 = 7.86869 13114 56132 59100D-4,
     *             F5 = 1.84631 83175 10054 68180D-5,
     *             F6 = 1.42151 17583 16445 88870D-7,
     *             F7 = 2.04426 31033 89939 78564D-15)
C       HASH SUM EF    47.52583 31754 92896 71629
C
        IFAULT = 0
        Q = P - HALF
        IF (ABS(Q) .LE. SPLIT1) THEN
          R = CONST1 - Q * Q
          PPND16 = Q * (((((((A7 * R + A6) * R + A5) * R + A4) * R + A3)
     *                  * R + A2) * R + A1) * R + A0) /
     *                (((((((B7 * R + B6) * R + B5) * R + B4) * R + B3)
     *                  * R + B2) * R + B1) * R + ONE)
          RETURN
        ELSE
          IF (Q .LT. ZERO) THEN
            R = P
          ELSE
            R = ONE - P
          END IF
          IF (R .LE. ZERO) THEN
            IFAULT = 1
            PPND16 = ZERO
            RETURN
          END IF
          R = SQRT(-LOG(R))
          IF (R .LE. SPLIT2) THEN
            R = R - CONST2
            PPND16 = (((((((C7 * R + C6) * R + C5) * R + C4) * R + C3)
     *                  * R + C2) * R + C1) * R + C0) /
     *               (((((((D7 * R + D6) * R + D5) * R + D4) * R + D3)
     *                  * R + D2) * R + D1) * R + ONE)
          ELSE
            R = R - SPLIT2
            PPND16 = (((((((E7 * R + E6) * R + E5) * R + E4) * R + E3)
     *                  * R + E2) * R + E1) * R + E0) /
     *               (((((((F7 * R + F6) * R + F5) * R + F4) * R + F3)
     *                  * R + F2) * R + F1) * R + ONE)
          END IF
          IF (Q .LT. ZERO) PPND16 = - PPND16
          RETURN
        END IF
        END


	real*8 function uniform()
c
c	Generate uniformly distributed random numbers using the 32-bit
c	generator from figure 3 of:
c	L'Ecuyer, P. Efficient and portable combined random number
c	generators, C.A.C.M., vol. 31, 742-749 & 774-?, June 1988.
c
c	The cycle length is claimed to be 2.30584E+18
c
c	Seeds can be set by calling the routine set_uniform
c
c	It is assumed that the Fortran compiler supports long variable
c	names, and integer*4.
c
	integer*4 z, k, s1, s2
	common /unif_seeds/ s1, s2
	save /unif_seeds/
c
	k = s1 / 53668
	s1 = 40014 * (s1 - k * 53668) - k * 12211
	if (s1 .lt. 0) s1 = s1 + 2147483563
c
	k = s2 / 52774
	s2 = 40692 * (s2 - k * 52774) - k * 3791
	if (s2 .lt. 0) s2 = s2 + 2147483399
c
	z = s1 - s2
	if (z .lt. 1) z = z + 2147483562
c
	uniform = z / 2147483563.
	return
	end


	subroutine set_uniform(seed1, seed2)
c
c	Set seeds for the uniform random number generator.
c
	integer*4 s1, s2, seed1, seed2
	common /unif_seeds/ s1, s2
	save /unif_seeds/

	s1 = seed1
	s2 = seed2
	return
	end







